// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

/// # Expression manipulation core library
///
/// ## Conventions
/// Argument counts for two expressions that have the same function or pattern
/// ID are supposed to be the same. When unchecked expressions from an external
/// input source are used the sanity of the argument counts should be checked
/// first.
///
/// ## Core tasks
///
/// 1. Find all differences between two expressions.
/// 2. Be able to resolve each difference by one of these methods:
///    - Resolve the difference using numeric evaluation
///    - Resolve the difference using numeric expansion
///    - Rearrange arguments of chained/nested rearrangeable functions
///    - Redistribute expression terms according to distributive properties
///    - Match against a valid difference pattern (formula)
///    - Break difference into smaller parts and resolve each
/// 3. Be able to apply evaluation, rearrangement, redistribution, substitution.
library expr;

import 'dart:typed_data';

import 'package:quiver/core.dart';
import 'package:quiver/iterables.dart';
import 'package:quiver/collection.dart';

import 'src/interface/reader.dart';

part 'src/core/expr.dart';
part 'src/core/evaluate.dart';
part 'src/core/transform.dart';
part 'src/core/mapping.dart';
part 'src/core/pattern.dart';
part 'src/core/map.dart';
part 'src/core/compare.dart';
part 'src/core/rearrange.dart';
part 'src/core/redistribute.dart';
part 'src/core/substitute.dart';

part 'src/interface/codec.dart';
part 'src/interface/operator.dart';
part 'src/interface/parser.dart';
part 'src/interface/template.dart';
part 'src/interface/printer.dart';

part 'src/lemma/structs.dart';
part 'src/lemma/rewrite.dart';
part 'src/lemma/lemma.dart';
part 'src/lemma/split.dart';

part 'src/misc/detect.dart';
