// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of expr;

/// Compare [input] transform with [pattern] transform. Returns true if the
/// pattern matches. Be careful with using a pre-existing [mapping], this may
/// have unintended side-effects when it conflicts with the source mapping.
bool compareTransforms(
    Transform input, Transform pattern, Set<int> canEval, EvalExpr eval,
    {Mapping mapping, Set<int> evaluateBefore, Set<int> evaluateAfter}) {
  final m = mapping ?? new Mapping();

  // Get source mapping.
  final t1 = new Transform(pattern.source, input.source);
  final sourceMapping = findSmallestTransforms(t1);
  mergeMappings(m, sourceMapping);

  // Try to normalize mapping through evaluation.
  for (final source in m.keys) {
    final targets = m[source];
    if (targets.length > 1) {
      // Evaluate all to simpler expressions.
      // Note that this approach does not fit all cases. E.g. if the targets are
      // `1 + (2 + 3)` and `1 + 5`, then both will be simplified to 6. It is not
      // possible to just simplify to `1 + 5`.
      m[source] = targets.map((expr) {
        final evaluate = new Set<int>();
        final result = tryEvaluateExpr(expr, canEval, eval, evaluate);

        if (evaluateBefore != null) {
          evaluateBefore.addAll(evaluate.map((i) => i + expr.offset));
        }

        return result;
      }).toSet();
    }
  }

  // Terminate if there are still unnormalized transforms.
  // This means that evaluation alone must be enough to normalized the source
  // mapping.
  if (isNormalizedMapping(m)) {
    final pMapping = getPatternMapping(m);

    // The source mapping must be eliminated to prove the pattern source fits.
    // This is necessary on top of normalization because an initial mapping can
    // be passed, and the source mapping might contain non-patterns.
    // For example, apply `#a + #b + (#a + #b)` to `1 + 2 + 3` so that
    // `#a + #b => 3` can be eliminated using the found pattern mapping and
    // evaluation.
    // Target evaluation (in this case evaluation of the input expression
    // source) must be stored in [evaluateBefore] because evaluation is needed
    // in order to apply a straightforward substitution.
    final sourceFits = m.every((source, targets) {
      final t = new Transform(source, targets.single);
      return _canEliminate(t, canEval, eval,
          evaluateTarget: evaluateBefore, mapping: pMapping);
    });

    // Construct expected target using the created pattern mapping.
    final patternTarget = mapExpr(pattern.target, pMapping);
    final t2 = new Transform(patternTarget, input.target);
    final targetMapping = findSmallestTransforms(t2);

    // Terminate if the target mapping is not normalized. The multiple mapping
    // targets would require extra steps after substitution which it outside the
    // scope of this algorithm.
    if (isNormalizedMapping(targetMapping)) {
      // The target mapping must be eliminated to prove the pattern target fits.
      final targetFits = targetMapping.every((source, targets) {
        final t = new Transform(source, targets.single);
        return _canEliminate(t, canEval, eval,
            evaluateSource: evaluateAfter, mapping: pMapping);
      });

      return sourceFits && targetFits;
    } else {
      return false;
    }
  } else {
    return false;
  }
}

/// Test if transform [t] can be reduced to an identity transform by rewriting
/// using the given [mapping] and evaluating using [eval]. Evaluation indices
/// will be merged with [evaluateSource] and [evaluateTarget] if the transform
/// can indeed be eliminated. Note that these indices are also offset by the
/// offsets defined in the input transform. This is a helper function that
/// should only be used internally.
bool _canEliminate(Transform t, Set<int> canEval, EvalExpr eval,
    {Mapping mapping, Set<int> evaluateSource, Set<int> evaluateTarget}) {
  final evaluate1 = new Set<int>();
  final evaluate2 = new Set<int>();
  final t2 = mapping != null ? mapTransform(t, mapping) : t;
  final t3 = tryEvaluateTransform(t2, canEval, eval, evaluate1, evaluate2);
  if (isIdentityTransform(t3)) {
    if (evaluateSource != null) {
      evaluateSource.addAll(evaluate1.map((i) => i + t2.source.offset));
    }
    if (evaluateTarget != null) {
      evaluateTarget.addAll(evaluate2.map((i) => i + t2.target.offset));
    }
    return true;
  } else {
    return false;
  }
}
