// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of expr;

/// Evaluation function, should evaluate the result of the computation
/// associated with [functionId] with the given [arguments].
typedef num EvalExpr(int functionId, List<num> arguments);

/// Evaluate the given [input] expression. Returns the resulting numeric value.
/// This makes the assumption that the [input] expression can be fully
/// evaluated. If the input is just a single number its value will be returned.
num evaluateExpr(Expr input, EvalExpr eval) {
  if (input.first.type == ExprType.number) {
    return input.first.value;
  }

  final funStack = new List<ExprObject>();
  final argStack = new List<List<num>>();

  for (var i = 0; i < input.length; i++) {
    final object = input[i];
    switch (object.type) {
      case ExprType.number:
        argStack.last.add(object.value);
        break;

      case ExprType.function:
        funStack.add(object);
        argStack.add(new List<num>());
        break;

      case ExprType.pattern:
        throw new ArgumentError('input cannot contain patterns');
    }

    while (argStack.last.length == funStack.last.argumentCount) {
      final int functionId = funStack.removeLast().value;
      final evalValue = eval(functionId, argStack.removeLast());
      if (argStack.isEmpty) {
        return evalValue;
      } else {
        argStack.last.add(evalValue);
      }
    }
  }

  throw new ArgumentError('incomplete input expression');
}

/// Evaluate the given [input] expression at the given [index] using [eval].
Expr applyEvaluation(Expr input, int index, EvalExpr eval) {
  final subExpr = getSubExpr(input, index);
  final value = evaluateExpr(subExpr, eval);
  final replacement = new Expr([new ExprObject(ExprType.number, value, 0)]);
  return replaceSubExpr(input, index, replacement);
}

/// Try to evaluate as much of the given [input] expression as possible and
/// return the resulting expression. This implementation can probably be
/// implemented more efficiently with a customized approach. The evaluated
/// sub-expression indices are stored in [evaluated].
Expr tryEvaluateExpr(Expr input, Set<int> canEval, EvalExpr eval,
    [Set<int> evaluated]) {
  final _evaluated = evaluated ?? new Set<int>();
  final exprContent = new List<ExprObject>();
  for (var i = 0; i < input.length;) {
    final subExpr = getSubExpr(input, i);
    final funIds = getExprTypeValues(subExpr, ExprType.function);
    final patIds = getExprTypeValues(subExpr, ExprType.pattern);
    if (patIds.isEmpty && funIds.isNotEmpty && canEval.containsAll(funIds)) {
      // Sub-expression can be evaluated, so evaluate it and add the index to
      // the output.
      _evaluated.add(i);
      final value = evaluateExpr(subExpr, eval);
      exprContent.add(new ExprObject(ExprType.number, value, 0));
      i += subExpr.length;
    } else {
      // Add object copy to output and proceed.
      exprContent.add(input[i]);
      i++;
    }
  }
  return new Expr(exprContent, input.offset);
}
