// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of expr;

/// This class is supposed to be used as immutable class. If the content is
/// modified a copy of all contents is returned.
class Expr extends DelegatingList<ExprObject> {
  /// Note that the offset is not considered when comparing expressions, it is
  /// just a piece of metadata that can be useful in some situations.
  ///
  /// The use of this field is somewhat non-transparent and has some serious
  /// downsides. It is used because is provides overall data-structure
  /// simplification.
  ///
  /// Generally speaking algorithms should not have anything to do with this
  /// field. It should not be used when addressing an expression index but it
  /// should be copied when creating a new expression from an existing one.
  final int offset;

  /// LaTeX string belonging to this expression. This is metadata that is part
  /// of the QEDb schema and does not have any additional role in this library.
  /// By convention there is no obligation to retain this value during
  /// processing.
  String latex;

  @override
  final List<ExprObject> delegate;

  Expr(List<ExprObject> content, [this.offset = 0])
      : delegate = new List<ExprObject>.from(content, growable: false);

  factory Expr.copy(Expr src) => new Expr(src, src.offset);

  @override
  int get hashCode => computeExprHash(this);

  @override
  bool operator ==(other) {
    return other is Expr && compareExprs(this, other);
  }
}

/// There are three types of expression objects
enum ExprType { number, function, pattern }

/// Expression object
/// Note that since this class is immutable, it is not necessary to ever create
/// deep copies of this object.
class ExprObject {
  final ExprType type;
  final num value;
  final int argumentCount;

  ExprObject(this.type, this.value, this.argumentCount);

  @override
  int get hashCode => computeExprObjectHash(this);

  @override
  bool operator ==(other) {
    return other is ExprObject && compareExprObjects(this, other);
  }
}

/// Compute hash code for [input] object.
int computeExprObjectHash(ExprObject input) {
  return hash3(input.type, input.value, input.argumentCount);
}

/// Compute hash code for [input] expression.
int computeExprHash(Expr input) {
  return hashObjects(input);
}

/// Test if [a] and [b] are equivalent objects. Does not explicitly check if the
/// argument count matches too. This is assumed and should be checked
/// separately.
bool compareExprObjects(ExprObject a, ExprObject b) {
  final equivalent = a.type == b.type && a.value == b.value;

  // Make sure expression objects are sane.
  if (equivalent && a.argumentCount != b.argumentCount) {
    throw new ArgumentError('incompatible argument count between a and b');
  }

  return equivalent;
}

/// Test if [a] and [b] are fully equivalent expressions. Note that this does
/// NOT take expression offsets into account.
bool compareExprs(Expr a, Expr b) {
  if (a.length != b.length) {
    return false;
  }
  for (var i = 0; i < a.length; i++) {
    if (!compareExprObjects(a[i], b[i])) {
      return false;
    }
  }
  return true;
}

/// Skip object content for [n] objects starting at [startIndex] and return the
/// resulting offset. This is done by summing all argument counts + [n] until
/// this is equal to the offset.
int skipObjectContent(Expr input, int startIndex, int n) {
  assert(n >= 0);
  var offset = 0, sum = n;
  while (offset < sum) {
    sum += input[startIndex + offset].argumentCount;
    offset++;
  }
  return offset;
}

/// Create new expression by copying the object and its contents from the
/// [input] expression at [index].
Expr getSubExpr(Expr input, int index) {
  final length = skipObjectContent(input, index, 1);
  return new Expr(input.sublist(index, index + length), input.offset + index);
}

/// Return the set of all values of objects with the given [type] in the given
/// [input] expression.
Set<int> getExprTypeValues(Expr input, ExprType type) {
  return input
      .where((object) => object.type == type)
      .map((object) => object.value.toInt())
      .toSet();
}

/// Get all function arguments for the given [input] expression.
List<Expr> getExprArguments(Expr input) {
  final arguments = new List<Expr>(input.first.argumentCount);
  for (var i = 1, c = 0; i < input.length && c < arguments.length; c++) {
    final subExpr = getSubExpr(input, i);
    arguments[c] = subExpr;
    i += subExpr.length;
  }
  return arguments;
}
