// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of expr;

/// Rewrite [input] expression by applying all transforms in [mapping]. This
/// algorithm could be considered rather inefficient, however it is also
/// flexible and complete.
Expr mapExpr(Expr input, Mapping mapping) {
  final exprContent = new List<ExprObject>();

  OUTER:
  for (var i = 0; i < input.length;) {
    final subExpr = getSubExpr(input, i);

    // Literal match.
    if (mapping.containsKey(subExpr)) {
      // Copy substitute expression.
      exprContent.addAll(mapping[subExpr].single);
      i += subExpr.length;
    } else {
      // Matching a pattern with arguments is more complicated.
      final topObj = input[i];
      if (topObj.type == ExprType.pattern && topObj.argumentCount > 0) {
        // Try to find normalized mapping for this pattern.
        for (final from in mapping.keys) {
          if (compareExprObjects(topObj, from.first)) {
            // Find transforms between pattern in [from] and pattern in [input].
            final t = new Transform(from, subExpr);
            final m1 = findSmallestTransforms(t);

            // Rewrite [m1] using mapping available in [mapping].
            final m2 = mapMapping(m1, mapping);

            // Apply rewritten mapping to the pattern target.
            if (isNormalizedMapping(m2)) {
              final newSubExpr = mapExpr(mapping[from].single, m2);
              exprContent.addAll(newSubExpr);
            } else {
              throw new Exception('conflict in pattern to pattern mapping');
            }

            i += subExpr.length;
            continue OUTER;
          }
        }
      }

      // Fallback to processing next object.
      exprContent.add(input[i]);
      i++;
    }
  }
  return new Expr(exprContent, input.offset);
}

/// Rewrite given transform [t] using [mapExpr] and the provided [mapping].
Transform mapTransform(Transform t, Mapping mapping) {
  final source = mapExpr(t.source, mapping);
  final target = mapExpr(t.target, mapping);
  return new Transform(source, target);
}

/// Rewrite given mapping [m] using [mapExpr] and the provided [mapping].
Mapping mapMapping(Mapping m, Mapping mapping) {
  final out = new Mapping();
  m.forEach((source, targets) {
    final newSource = mapExpr(source, mapping);
    final newTargets = targets.map((expr) => mapExpr(expr, mapping));
    out.put(newSource);
    out[newSource].addAll(newTargets);
  });
  return out;
}

/// Replace the sub-expression in [input] at [index] with [replacement].
Expr replaceSubExpr(Expr input, int index, Expr replacement) {
  final skipOffset = skipObjectContent(input, index, 1);
  final exprContent = new List<ExprObject>();
  exprContent.addAll(input.sublist(0, index));
  exprContent.addAll(replacement);
  exprContent.addAll(input.sublist(index + skipOffset));
  return new Expr(exprContent, input.offset);
}
