// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of expr;

/// The one class to rule them all.
class Mapping extends DelegatingMap<Expr, Set<Expr>> {
  @override
  final delegate = new Map<Expr, Set<Expr>>();

  /// Shorthand for [putIfAbsent] for [key].
  void put(Expr key) {
    putIfAbsent(key, () => new Set<Expr>());
  }

  /// Add the given transform [t].
  void add(Transform t) {
    put(t.source);
    this[t.source].add(t.target);
  }

  /// Test if [condition] is true for all pairs.
  bool every(bool condition(Expr source, Set<Expr> targets)) {
    for (final source in keys) {
      if (!condition(source, this[source])) {
        return false;
      }
    }
    return true;
  }
}

/// Test if each transform in mapping [m] has only one target.
bool isNormalizedMapping(Mapping m) {
  return m.values.every((targets) => targets.length == 1);
}

/// Merge mapping [src] with the given mapping [dst].
void mergeMappings(Mapping dst, Mapping src) {
  for (final source in src.keys) {
    dst.put(source);
    dst[source].addAll(src[source]);
  }
}

/// Break the given transform [t] into a list of smallest possible transforms.
Mapping findSmallestTransforms(Transform t) {
  final mapping = new Mapping();
  for (var fi = 0, ti = 0; fi < t.source.length;) {
    // If the input expressions are broken a range error will be thrown, we're
    // fine with that.
    final sourceObj = t.source[fi];
    final targetObj = t.target[ti];

    // Compare the two nodes.
    if (!compareExprObjects(sourceObj, targetObj)) {
      // The nodes are different, so we can add them to the mapping. While we do
      // this we also skip the node arguments.
      final source = getSubExpr(t.source, fi);
      final target = getSubExpr(t.target, ti);
      fi += source.length;
      ti += target.length;
      mapping.add(new Transform(source, target));
    } else {
      // Else: The nodes are equal.
      // Compare the node arguments. Since we are comparing against the entire
      // input expression, we can simply keep iterating forward.
      //
      // Note that if t.source has more arguments than t.target despite having
      // the same type and value, this function will not actually fail even
      // though the result is misaligned. However an error is already thrown in
      // [compareExprObjects] to prevent this. No further action is required.
      fi++;
      ti++;
    }
  }

  return mapping;
}

/// If the given transform has a common top-level function or pattern ID, break
/// it into a single transform per argument. Else an empty list is returned
List<Transform> getArgumentTransforms(Transform t) {
  final sourceObj = t.source.first;
  final targetObj = t.target.first;
  if (sourceObj.type != ExprType.number &&
      compareExprObjects(sourceObj, targetObj)) {
    final arguments = new List<Transform>();
    for (var i = 0, fi = 1, ti = 1; i < sourceObj.argumentCount; i++) {
      final source = getSubExpr(t.source, fi);
      final target = getSubExpr(t.target, ti);
      fi += source.length;
      ti += target.length;
      arguments.add(new Transform(source, target));
    }
    return arguments;
  } else {
    return [];
  }
}
