// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of expr;

/// Rearrangement data
class Rearrangement {
  final int index;
  final List<int> format;

  Rearrangement(this.index, this.format);
}

/// Find top-level rearrangements in the given transform. If no rearrangements
/// can be found, the returned list if empty. Nested rearrangements are handled
/// first.
List<Rearrangement> findRearrangements(Transform t, Set<int> canRearrange) {
  final sourceObj = t.source.first;
  final targetObj = t.target.first;

  // Check if transform top level function is rearrangeable.
  if (sourceObj.type == ExprType.function &&
      compareExprObjects(sourceObj, targetObj) &&
      canRearrange.contains(sourceObj.value)) {
    // Collect arguments.
    final functionId = sourceObj.value;
    final tFormat = new List<int>();
    final sArgs = collectNestedArguments(t.source, functionId);
    final tArgs = collectNestedArguments(t.target, functionId, format: tFormat);

    // Check if arguments are comparable. Note that arguments might itself have
    // been rearranged. To take this into account, we use a special hash.
    int toHash(Expr a) => computeRearrangeableHash(a, canRearrange);
    final fHashes = sArgs.map(toHash).toList();
    final tHashes = tArgs.map(toHash).toList();

    // To compare both lists we need to make a copy and sort it, like in
    // [computeRearrangeableHash].
    final fHashesSorted = fHashes.toList()..sort();
    final tHashesSorted = tHashes.toList()..sort();
    final comparable = listsEqual(fHashesSorted, tHashesSorted);
    if (comparable) {
      // Start building rearrangement stack.
      final out = new List<Rearrangement>();
      final used = new List<bool>.filled(sArgs.length, false);

      // Iterate through format and replace indices.
      for (var i = 0; i < tFormat.length; i++) {
        if (tFormat[i] != -1) {
          // Find first matching argument that is not yet used. This is supposed
          // to always result in a valid index at this point.
          var sourceIndex = 0;
          final targetIndex = tFormat[i];
          while (fHashes[sourceIndex] != tHashes[targetIndex] ||
              used[sourceIndex]) {
            sourceIndex++;
          }
          tFormat[i] = sourceIndex;
          used[sourceIndex] = true;

          // If these arguments are equal but the expressions are not, there is
          // supposed to be a nested rearrangement.
          if (sArgs[sourceIndex] != tArgs[targetIndex]) {
            final source = sArgs[sourceIndex], target = tArgs[targetIndex];
            final subTransform = new Transform(source, target);
            out.addAll(findRearrangements(subTransform, canRearrange));
          }
        }
      }

      out.add(new Rearrangement(t.source.offset, tFormat));
      return out;
    } else {
      return [];
    }
  } else {
    return [];
  }
}

/// Compute hash for the given [input] expression that is unique for any
/// arrangement of its rearrangeable children. Recursion seems the best choice
/// for this function.
int computeRearrangeableHash(Expr input, Set<int> canRearrange) {
  final topObj = input.first;
  if (topObj.type == ExprType.function && canRearrange.contains(topObj.value)) {
    final nestedArguments = collectNestedArguments(input, topObj.value);
    final hashes = nestedArguments
        .map((argument) => computeRearrangeableHash(argument, canRearrange))
        .toList();
    hashes.sort();
    return hashObjects(hashes);
  } else {
    return computeExprHash(input);
  }
}

/// Collect list of arguments that are nested under the same [functionId].
/// The arguments contain references to the expression objects.
/// [format]: If provided, a format pattern will be written to this.
List<Expr> collectNestedArguments(Expr input, int functionId,
    {List<int> format}) {
  final out = new List<Expr>();
  for (var i = 0; i < input.length;) {
    final object = input[i];

    // If object is a function with the given function ID, continue to look at
    // all its children. If not, copy it to the output list.
    if (object.type == ExprType.function && object.value == functionId) {
      if (format != null) {
        format.add(-1);
      }
      i++;
    } else {
      if (format != null) {
        format.add(out.length);
      }

      out.add(getSubExpr(input, i));
      i += out.last.length;
    }
  }
  return out;
}

/// Apply rearrangement specified by [r] to the given [input] expression.
/// The function ID of the first function is used.
Expr rearrangeExpr(Expr input, Rearrangement r) {
  final subExpr = getSubExpr(input, r.index);
  final topObj = subExpr.first;
  if (topObj.type != ExprType.function) {
    throw new ArgumentError('rearrangement target is not a function');
  }

  // Assemble rearranged expression.
  final functionId = topObj.value;
  final argumentCount = topObj.argumentCount;
  final arguments = collectNestedArguments(subExpr, functionId);
  final result = new List<ExprObject>();
  for (final i in r.format) {
    if (i == -1) {
      result.add(new ExprObject(ExprType.function, functionId, argumentCount));
    } else {
      result.addAll(arguments[i]);
    }
  }

  // Return substituted result.
  return replaceSubExpr(input, r.index, new Expr(result));
}
