// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of expr;

/// Algorithm to find redistribution of arguments in argument trees that are a
/// combination of two functions f and g that have the following property:
/// `f(g(a, b)) = g(f(a), f(b))`.
///
/// For example, given `f(x) = w * x` and `g(x, y) = x + y`
/// Then `f(g(a, b)) = w * (a + b) = w * a + w * b = g(f(a), f(b))`.
///
/// Function property information is contained in [canRedistribute] as a map
/// between the f/g function ID and the list of partial application gap argument
/// indices of the f function (it is up to the caller to sort this list and make
/// sure elements are unique).
Redistribution findRedistribution(Transform t, Set<int> canRearrange,
    Map<DistFnKey, List<int>> canRedistribute) {
  final key = _findDistFnKey(t.source, canRedistribute.keys.toSet());
  if (key == null) {
    return null;
  }

  // Check that g is rearrangeable (the algorithm assumes this by shuffling the
  // unwinded expressions).
  if (!canRearrange.contains(key.g)) {
    return null;
  }

  // Unwind source and target.
  final fGapArgs = canRedistribute[key];
  final source = unwindExprDist(t.source, key, fGapArgs);
  final target = unwindExprDist(t.target, key, fGapArgs);

  // Construct redistribution tree.
  final sourceStack = flattenMergeTree(source);
  final used = new List<bool>.filled(sourceStack.length, false);
  final rootNode = _convertDistMerge(target, sourceStack, used, canRearrange);

  // Check if redistribution is complete.
  if (rootNode == null || !used.every((u) => u)) {
    return null;
  } else {
    return new Redistribution(key, fGapArgs, rootNode);
  }
}

/// Apply redistribution defined by [r] to [input].
Expr redistributeExpr(Expr input, Redistribution r) {
  final source = flattenMergeTree(unwindExprDist(input, r.key, r.fGapArgs));
  final result = _computeRedistribution(r.key, r.node, source);
  return new Expr(result, input.offset); // Offsets are messed up
}

Expr _computeRedistribution(DistFnKey k, RedistrNode node, List<Expr> source) {
  if (node.sourceIndex != -1) {
    final expr = source[node.sourceIndex];
    return node.r != null ? rearrangeExpr(expr, node.r) : expr;
  } else {
    final a = _computeRedistribution(k, node.a, source);
    final b = _computeRedistribution(k, node.b, source);
    return mergeExpressions(k.f, k.g, a, b, node.mergePoint);
  }
}

/// Merge two expressions at the given [position] (descending list of argument
/// indices) by wrapping sub-expressions at this point in [g] function. Also
/// checks that all parents are the [f] function.
Expr mergeExpressions(int f, int g, Expr a, Expr b, List<int> position) {
  // Iterate to targeted sub-expression in a and b.
  var ai = 0, bi = 0;
  for (var i in position) {
    if ((a[ai].type != ExprType.function || a[ai].value != f) ||
        (b[bi].type != ExprType.function || b[bi].value != f)) {
      throw new ArgumentError('invalid position');
    } else {
      ai++;
      bi++;
      while (i-- > 0) {
        ai += skipObjectContent(a, ai, 1);
        bi += skipObjectContent(b, bi, 1);
      }
    }
  }

  // Merge sub-expressions and substitute into a and b. Then check if the
  // results are the same.
  final mergedContent = new List<ExprObject>();
  mergedContent.add(new ExprObject(ExprType.function, g, 2));
  mergedContent.addAll(getSubExpr(a, ai));
  mergedContent.addAll(getSubExpr(b, bi));
  final merged = new Expr(mergedContent);

  final aa = replaceSubExpr(a, ai, merged);
  final bb = replaceSubExpr(b, bi, merged);
  if (aa == bb) {
    return aa;
  } else {
    throw new ArgumentError('asymmetric merge');
  }
}

/// Unwind [input] expression distribution using the given [key].
DistMerge unwindExprDist(Expr input, DistFnKey key, List<int> fGapArgs) {
  if (input.first.type == ExprType.function) {
    final id = input.first.value;
    final args = getExprArguments(input);

    if (id == key.f) {
      // Expand all partial application gap arguments.
      final mergeSource = new DistMerge(input, null, null, null);
      return fGapArgs.fold(mergeSource, (node, i) {
        final argNode = unwindExprDist(args[i], key, fGapArgs);
        return _appendMerges(argNode, node, i);
      });
    } else if (id == key.g) {
      // For now we basically expect g to be a binary operator which has quite
      // easy to understand merge behaviour.
      if (args.length != 2) {
        throw new ArgumentError('illegal g function argument count');
      }

      // Pack into a single merge.
      final ca = unwindExprDist(args[0], key, fGapArgs);
      final cb = unwindExprDist(args[1], key, fGapArgs);
      return new DistMerge(null, ca, cb, []);
    }
  }

  // Fallback (expression cannot be split up further)
  return new DistMerge(input, null, null, null);
}

/// Append merge [b] to the ends of merge tree [a]. This means that a copy of
/// merge tree [b] is added to the ends of merge tree [a] and the expressions at
/// the ends of the old merge tree [a] are substituted in the expressions at the
/// ends of the new merge tree (after [b] was appended). A recursive approach is
/// used.
DistMerge _appendMerges(DistMerge a, DistMerge b, int i) {
  if (a.src != null) {
    if (b.src != null) {
      final args = getExprArguments(b.src);
      final offset = 1 + args.sublist(0, i).fold(0, (int a, b) => a + b.length);
      final expr = replaceSubExpr(b.src, offset, a.src);
      return new DistMerge(expr, null, null, null);
    } else {
      final ba = _appendMerges(a, b.a, i);
      final bb = _appendMerges(a, b.b, i);
      return new DistMerge(null, ba, bb, b.position);
    }
  } else {
    final position = new List<int>.from(a.position);
    position.insert(0, i);
    final aa = _appendMerges(a.a, b, i);
    final ab = _appendMerges(a.b, b, i);
    return new DistMerge(null, aa, ab, position);
  }
}

/// Flatten merge tree.
List<Expr> flattenMergeTree(DistMerge merge) {
  if (merge.src != null) {
    // Reset expression offset to 0.
    return [new Expr(merge.src)];
  } else {
    final list = new List<Expr>();
    list.addAll(flattenMergeTree(merge.a));
    list.addAll(flattenMergeTree(merge.b));
    return list;
  }
}

RedistrNode _convertDistMerge(
    DistMerge node, List<Expr> source, List<bool> used, Set<int> canRearrange) {
  assert(source.length == used.length);
  if (node.src != null) {
    // Search for expression in source.
    for (var i = 0; i < source.length; i++) {
      if (!used[i]) {
        // Note that we have to reset the expression offset of [node.src] to 0.
        final t = new Transform(new Expr(node.src), source[i]);
        if (isIdentityTransform(t)) {
          used[i] = true;
          return new RedistrNode(i, null, null, null, null);
        } else {
          // Note that with the current implementation it only makes sense for
          // a single rearrangement to exist.
          final rearrange = findRearrangements(t, canRearrange);
          if (rearrange.isNotEmpty) {
            used[i] = true;
            return new RedistrNode(i, rearrange.single, null, null, null);
          }
        }
      }
    }

    // Fallback
    return null;
  } else {
    final p = node.position;
    final a = _convertDistMerge(node.a, source, used, canRearrange);
    final b = _convertDistMerge(node.b, source, used, canRearrange);
    return a == null || b == null ? null : new RedistrNode(-1, null, a, b, p);
  }
}

/// Find valid IDs for f and g function.
DistFnKey _findDistFnKey(Expr input, Set<DistFnKey> keys) {
  if (input.first.type != ExprType.function) {
    return null;
  }

  final int id1 = input.first.value;
  for (var i = 1; i < input.length;) {
    if (input[i].type == ExprType.function) {
      final int id2 = input[i].value;
      final key1 = new DistFnKey(id1, id2);
      final key2 = new DistFnKey(id2, id1);
      if (keys.contains(key1)) {
        return key1;
      } else if (keys.contains(key2)) {
        return key2;
      }
    }
    // Iterate through arguments.
    i += skipObjectContent(input, i, 1);
  }
  return null;
}

/// Redistribution information
class Redistribution {
  final DistFnKey key;
  final List<int> fGapArgs;
  final RedistrNode node;

  Redistribution(this.key, this.fGapArgs, this.node);
}

/// Node in redistribution tree (converted into [DistMerge])
class RedistrNode {
  final int sourceIndex;
  final Rearrangement r;
  final RedistrNode a, b;
  final List<int> mergePoint;

  RedistrNode(this.sourceIndex, this.r, this.a, this.b, this.mergePoint);
}

/// ID of the f and g function.
class DistFnKey {
  final int f, g;
  DistFnKey(this.f, this.g);

  @override
  int get hashCode => hash2(f, g);

  @override
  bool operator ==(x) => x is DistFnKey && x.f == f && x.g == g;
}

/// Distribution merge point
///
/// The limit of just two merge targets ([a] and [b]) is intentional because it
/// makes it easier to reason about the problem, and fits the use case.
class DistMerge {
  final Expr src;
  final DistMerge a, b;
  final List<int> position;

  DistMerge(this.src, this.a, this.b, this.position) {
    assert(src == null || (a == null && b == null && position == null));
  }
}
