// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of expr;

/// Encode [input] expression as slightly compressed binary buffer. We avoid
/// using Huffman tables here to make the codec simple to implement.
ByteBuffer encodeExprAsBytes(Expr input) {
  // Collection tables.
  final float62l = new List<double>();
  final fId = new List<int>();
  final pId = new List<int>();
  final fArgc = new List<int>();
  final pArgc = new List<int>();
  final int8l = new List<int>();
  final objects = new List<int>();
  final objectIdOffset = new List<int>();

  // Populate tables.
  for (final object in input) {
    switch (object.type) {
      case ExprType.number:
        final intv = object.value.toInt();
        if (intv == object.value && intv >= -128 && intv <= 127) {
          objectIdOffset.add(3);
          final i = int8l.indexOf(intv);
          if (i == -1) {
            objects.add(int8l.length);
            int8l.add(intv);
          } else {
            objects.add(i);
          }
        } else {
          objectIdOffset.add(0);
          final i = float62l.indexOf(object.value);
          if (i == -1) {
            objects.add(float62l.length);
            float62l.add(object.value.toDouble());
          } else {
            objects.add(i);
          }
        }
        break;

      case ExprType.function:
        objectIdOffset.add(1);
        final i = fId.indexOf(object.value);
        if (i == -1) {
          objects.add(fId.length);
          fId.add(object.value);
          fArgc.add(object.argumentCount);
        } else {
          objects.add(i);
        }
        break;

      case ExprType.pattern:
        objectIdOffset.add(2);
        final i = pId.indexOf(object.value);
        if (i == -1) {
          objects.add(pId.length);
          pId.add(object.value);
          pArgc.add(object.argumentCount);
        } else {
          objects.add(i);
        }
        break;
    }
  }

  // Offset structure indices.
  for (var i = 0; i < objects.length; i++) {
    var offset = objectIdOffset[i];
    objects[i] += --offset >= 0 ? float62l.length : 0;
    objects[i] += --offset >= 0 ? fId.length : 0;
    objects[i] += --offset >= 0 ? pId.length : 0;
  }

  // Concat function and pattern values.
  final id = concat([fId, pId]).toList();
  final argc = concat([fArgc, pArgc]).toList();

  // Compute total length.
  final maxLookupIndex = float62l.length + id.length + int8l.length;
  final headerSize = 8 +
      float62l.length * 8 +
      id.length * 4 +
      argc.length * 2 +
      int8l.length * 1;

  final bytesPerObject = maxLookupIndex > 255 ? 2 : 1;
  final headerPad = headerSize % bytesPerObject;
  final bufferSize = headerSize + headerPad + objects.length * bytesPerObject;
  final buffer = new Uint8List(bufferSize).buffer;
  var offset = 0;

  final dimView = new Uint16List.view(buffer, offset, 4);
  dimView.setAll(0, [float62l.length, fId.length, pId.length, int8l.length]);
  offset += dimView.lengthInBytes;
  final float64View = new Float64List.view(buffer, offset, float62l.length);
  float64View.setAll(0, float62l);
  offset += float64View.lengthInBytes;
  final idView = new Uint32List.view(buffer, offset, id.length);
  idView.setAll(0, id);
  offset += idView.lengthInBytes;
  final argcView = new Uint16List.view(buffer, offset, argc.length);
  argcView.setAll(0, argc);
  offset += argcView.lengthInBytes;
  final int8View = new Int8List.view(buffer, offset, int8l.length);
  int8View.setAll(0, int8l);
  offset += int8View.lengthInBytes + headerPad;

  // Write objects.
  if (bytesPerObject == 1) {
    final objectsView = new Uint8List.view(buffer, offset, objects.length);
    objectsView.setAll(0, objects);
  } else {
    final objectsView = new Uint16List.view(buffer, offset, objects.length);
    objectsView.setAll(0, objects);
  }

  return buffer;
}

/// Decode the byte data in [buffer] into an expression.
Expr decodeBytesIntoExpr(ByteBuffer buffer) {
  var offset = 0;

  // Read dimensions.
  final dimView = new Uint16List.view(buffer, offset, 4);
  final float64Count = dimView[0];
  final fIdCount = dimView[1];
  final pIdCount = dimView[2];
  final int8Count = dimView[3];
  offset += dimView.lengthInBytes;

  // Split buffer into parts.
  final float64View = new Float64List.view(buffer, offset, float64Count);
  offset += float64View.lengthInBytes;
  final idView = new Uint32List.view(buffer, offset, fIdCount + pIdCount);
  offset += idView.lengthInBytes;
  final argcView = new Uint16List.view(buffer, offset, fIdCount + pIdCount);
  offset += argcView.lengthInBytes;
  final int8View = new Int8List.view(buffer, offset, int8Count);
  offset += int8View.lengthInBytes;

  // Read objects.
  final objects = new List<ExprObject>();
  objects.addAll(float64View.map((v) => new ExprObject(ExprType.number, v, 0)));
  for (var i = 0; i < idView.length; i++) {
    final type = i < fIdCount ? ExprType.function : ExprType.pattern;
    objects.add(new ExprObject(type, idView[i], argcView[i]));
  }
  objects.addAll(int8View.map((v) => new ExprObject(ExprType.number, v, 0)));

  // Get object indices.
  final bytesPerObject = objects.length > 255 ? 2 : 1;
  offset += offset % bytesPerObject;
  final objectIndices = bytesPerObject == 1
      ? new Uint8List.view(buffer, offset)
      : new Uint16List.view(buffer, offset);

  // Assemble expression.
  return new Expr(objectIndices.map(objects.elementAt).toList());
}
