// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of expr;

/// Get function or pattern ID (indicated by [type]) the given string [label].
typedef int GetExprObjectId(ExprType type, String label);

/// Parse expression from [input] string. Uses the Shunting-yard algorithm.
/// Patterns are recognized by a preceeding [patternSymbol] which must be 1
/// character long. This is a one-pass, linear-time, linear-space algorithm.
///
/// The operators in [operators] should each have a unique symbol. In some cases
/// a prefix operator can be detected (e.g. `a*-b` or `a(-b)`). So it is allowed
/// to use the same symbol twice for a prefix and infix/postfix operator
/// (for example prefix and infix minus). This is not actively checked. If there
/// are operators with conflicting symbols (e.g. one symbol is the subset of the
/// other), the resulting symbol is undefined and the parser may fail.
///
/// A special case fo implicit multiplication is built into the parser:
/// Implicit multiplication: this is multiplication without a multiplication
/// symbol but implied by a missing infix operator. [imid] will be used as
/// temporary function ID for this operator. This is useful because alternative
/// operator properties may be specified for implicit multiplication.
/// [multiplyId] contains the actual function ID of multiplication, if it is
/// null, the parser will look for the asterisk symbol in the [operators] list.
Expr parseExpr(String input, List<Operator> operators, GetExprObjectId getId,
    [String patternSymbol = '#', int imid = 0, multiplyId]) {
  // Pattern symbol must be one character.
  final badPatternSymbols = ['(', ',', ')', ' '];
  if (patternSymbol.length != 1 || badPatternSymbols.contains(patternSymbol)) {
    throw new ArgumentError('illegal pattern symbol');
  }

  // Empty strings cannot be parsed.
  if (input.isEmpty) {
    throw const FormatException('input cannot be empty');
  }

  final output = new List<Expr>();
  final stack = new List<_StackElement>();
  final reader = new StringReader(input);

  // List of operator symbols.
  final operatorSymbols = new Set<String>();
  for (final operator in operators) {
    operatorSymbols.addAll(operator.symbols);
  }

  // Create list of all characters that are special and sign the end of a label.
  // This includes the first character of any operator symbol, the pattern
  // symbol, parentheses, a comma, and a whitespace.
  final specialChars = new List<int>();
  specialChars.addAll(operatorSymbols.map((sym) => sym.runes.first));
  specialChars.add(patternSymbol.runes.first);
  specialChars.addAll('(,) '.codeUnits);

  // Find multiply function ID.
  final mid = multiplyId ??
      operators.singleWhere((op) {
        return op.notation == OperatorNotation.infix &&
            op.symbols.contains('*');
      }).functionId;

  /// Tokens:
  /// - number
  /// - math operator
  /// - function name (or variable)
  /// - parentheses
  /// - argument separator

  // Initial operator distance = 2 because we did not pass any operator yet, now
  // the [_detectImplMul] function will at least not think we're close to one.
  var opDist = 2;

  // Distance in tokens between the beginning of the current block + 1
  // (whole equation, argument list argument, parentheses block)
  var blockDist = 1;

  while (!reader.eof) {
    reader.skipWhitespace();

    // Save pointer position before parsing the token.
    var startPtr = reader.ptr;

    // Try left parenthesis.
    if (reader.currentIs('(')) {
      _detectImplMul(opDist, blockDist, stack, output, operators, imid, mid);

      stack.add(new _StackElement(isLeftParenthesis: true));

      // Reset left block start distance.
      blockDist = 0;

      // Move to next token.
      reader.next();
    }

    // Try right parenthesis.
    else if (reader.currentIs(')')) {
      if (stack.isEmpty) {
        // The stack already ran out without finding a parentheses.
        throw const FormatException('mismatched parentheses');
      }

      // Pop all remaining stack elements.
      while (!stack.last.isLeftParenthesis) {
        _popStack(stack, output, imid, mid);
        if (stack.isEmpty) {
          // If the stack runs out without finding a left parentheses...
          throw const FormatException('mismatched parentheses');
        }
      }

      // Pop the left parenthesis.
      stack.removeLast();

      // If the last element in the stack is a function, pop it.
      if (stack.isNotEmpty && stack.last.isFunction) {
        _popStack(stack, output, imid, mid);
      }

      // Move to next token.
      reader.next();
    }

    // Try argument separator (comma).
    else if (reader.currentIs(',')) {
      // If stack is already empty, also throw an error.
      if (stack.isEmpty) {
        throw const FormatException('argument separator outside function');
      }

      // Pop operators.
      while (!stack.last.isLeftParenthesis) {
        _popStack(stack, output, imid, mid);
        if (stack.isEmpty) {
          throw const FormatException('argument separator outside function');
        }
      }

      // At the top of the stack there should now be a parenthesis, we can
      // increment the argc of the function before it.
      if (stack.length >= 2 && stack.last.isLeftParenthesis) {
        stack[stack.length - 2].argc++;
      } else {
        throw const FormatException('argument separators outside function');
      }

      // Rest block start distance.
      blockDist = 0;

      // Move to next token.
      reader.next();
    }

    // Try operator.
    else if (reader.matchAny(operatorSymbols)) {
      // Figure out if this might be a prefix operator.
      // * the previous token is also an operator
      // * this is the first token in the current block
      final prefix = opDist == 1 || blockDist == 1;

      // Find associated operator.
      final symbol = reader.str(startPtr, reader.ptr);
      final op = operators.singleWhere((op) =>
          op.symbols.contains(symbol) &&
          (op.notation == OperatorNotation.prefix) == prefix);

      // Add operator to stack.
      _stackAddOp(op, stack, output, operators, imid, mid);
      opDist = 0; // Reset operator distance.
    }

    // Else try to parse as number and fallback to functions.
    // Note that we have to check for reader.eof here.
    else if (!reader.eof) {
      _detectImplMul(opDist, blockDist, stack, output, operators, imid, mid);

      // Find number token.
      var nonDigits = 0; // Number of non-digits that were parsed.

      // Note that there can not be a negate sign here, since that would have
      // been parsed as minus operator, or as an unary minus.

      reader.skipDigits(); // Take any number of digits.

      // If this is a decimal point, skip it and any numbers following it.
      if (reader.currentIs('.')) {
        nonDigits++;
        reader.next();
        reader.skipDigits();
      }

      // The number is terminated here, check if a number was consumed:
      // The difference between the current pointer and the start position must
      // be at least one digit.
      if (reader.ptr - startPtr - nonDigits > 0) {
        // Create string from number segment, and parse.
        final value = num.parse(reader.str(startPtr, reader.ptr));

        // Push to output queue.
        output.add(new Expr([new ExprObject(ExprType.number, value, 0)]));

        // Note that we do NOT have to move the pointer here. This is because
        // by skipping digits, we have already moved past this token.
      } else {
        // Interpret as function.
        // If the current character is the pattern symbol, this is the start of
        // a pattern. We move startPtr forward to remove it from the token.
        final pattern = reader.currentIs(patternSymbol);
        if (pattern) {
          reader.next();
          startPtr++;
        }

        // Find label end.
        reader.skip((char) => !specialChars.contains(char));
        final label = reader.str(startPtr, reader.ptr);

        // Since we have implicit multiplication after whitespace or any other
        // characters, this is only a function if there is an opening
        // parenthesis directly after the name.
        var isSymbol = !reader.currentIs('(');
        if (!isSymbol) {
          // Move over opening parentheses.
          reader.next();

          // Check if maybe the token is still a symbol because it has 0 args.
          reader.skipWhitespace();
          if (reader.currentIs(')')) {
            isSymbol = true;
            reader.next();
          }
        }

        // Process token.
        final type = pattern ? ExprType.pattern : ExprType.function;
        final id = getId(type, label);
        if (isSymbol) {
          output.add(new Expr([new ExprObject(type, id, 0)]));
        } else {
          // We initially expect one argument, this will be incremented when we
          // find argument separators.
          stack.add(new _StackElement(id: id, argc: 1, isPattern: pattern));

          // Add the left parenthesis here.
          stack.add(new _StackElement(isLeftParenthesis: true));

          // Reset block start distance.
          blockDist = 0;
        }

        // We do NOT have to move the pointer here. By skipping characters
        // until we encountered a special character we are already at the next
        // token.
      }
    }

    opDist++;
    blockDist++;
  }

  // Drain stack.
  while (stack.isNotEmpty) {
    _popStack(stack, output, imid, mid);
  }

  // This should always be true. If there is a formatting error it will be
  // detected earlier.
  return output.single;
}

/// If there is a case of implicit multiplication, an implicit multiplication
/// operator is added to the stack. [opDist] is the distance to the last
/// operator in tokens. [blockStartDist] is the distance to the beginning of
/// the current expression block in tokens inclusive the current token. This
/// function is only called in appropriate situations, so the checks here are
/// not complete.
void _detectImplMul(int opDist, int blockStartDist, List<_StackElement> stack,
    List<Expr> output, List<Operator> operators, int imid, int mid) {
  // Implicit multiplication if all are true:
  // * This is not the first token in the current block
  // * The previous token is not an operator
  if (blockStartDist != 1 && opDist != 1) {
    final op = operators.singleWhere((op) => op.functionId == imid);
    _stackAddOp(op, stack, output, operators, imid, mid);

    // Note that we do NOT have to set the operator distance here. We
    // can set it to 1: since we have added an operator and are already
    // processing the next token, the distance is equal to 1.
    // However, this is not necessary since opDist is only used to check
    // if the distance is not equal to one, and in the next iteration
    // it will be incremented to 2. It is not possible for the operator
    // distance to be 0 at this point.
  }
}

/// Add operator to the [stack].
void _stackAddOp(Operator o1, List<_StackElement> stack, List<Expr> output,
    List<Operator> operators, int imid, int mid) {
  // While there is an operator o2 at the top of the stack, pop it if:
  // - o1 is left associative and its precedence is less or equal to o2
  // - o1 is right associative and its precedence is less than o2
  final assocOffset = o1.associativity == Associativity.ltr ? 1 : 0;
  while (stack.isNotEmpty && stack.last.isOperator) {
    final o2 = operators.singleWhere((op) => op.functionId == stack.last.id);
    if (o1.precedenceLevel < o2.precedenceLevel + assocOffset) {
      _popStack(stack, output, imid, mid);
    } else {
      break;
    }
  }

  // Add operator to stack.
  final argc = o1.notation == OperatorNotation.infix ? 2 : 1;
  stack.add(new _StackElement(id: o1.functionId, argc: argc, isOperator: true));
}

/// Pop and process element from the [stack].
/// This function should not be called to remove left parentheses.
/// (if there are still left parentheses left, there is a mismatch)
/// [imid]: implicit multiply function ID
/// [mid]: multiply function ID
void _popStack(
    List<_StackElement> stack, List<Expr> output, int imid, int mid) {
  final stackElement = stack.removeLast();

  // Check if arguments are in the stack.
  if (output.length < stackElement.argc) {
    throw const FormatException(
        'stack function arguments are not in the output queue');
  }

  // Get object type, ID, arguments.
  final type = stackElement.isPattern ? ExprType.pattern : ExprType.function;
  final id = stackElement.id == imid ? mid : stackElement.id;
  final arguments = new List<Expr>(stackElement.argc);
  for (var i = arguments.length - 1; i >= 0; i--) {
    arguments[i] = output.removeLast();
  }

  // Construct expression content.
  final exprContent = new List<ExprObject>();
  exprContent.add(new ExprObject(type, id, arguments.length));
  arguments.forEach(exprContent.addAll);

  output.add(new Expr(exprContent));
}

/// Element in parsing stack.
class _StackElement {
  final int id;
  int argc = 0;
  final bool isPattern, isOperator, isLeftParenthesis;

  _StackElement(
      {this.id,
      this.argc: 0,
      this.isPattern: false,
      this.isOperator: false,
      this.isLeftParenthesis: false});

  bool get isFunction => !isOperator && !isLeftParenthesis;
}
