// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of expr;

/// Provide printing template for given expression [object]. This is called for
/// all expression objects types including numbers.
typedef PrintTemplate GetPrintTemplate(ExprObject object);

/// Printing output sanitizer. When strings are appended in the printer that
/// are supposed to be separate but seem to belong together after appending,
/// this methods has the opportunity to return a String that has to be inserted
/// in between (such as a space). For example `sin(a)` should become `\sin a`
/// in LaTeX, not `\sina`).
typedef String InsertPrintSeparator(String a, String b);

/// Print the given [input] expression using printing templates provided by
/// [getTemplate]. It is required to provide a map between operator function ID
/// and their precedence level in [operators]. Usually [blockLeft] and
/// [blockRight] are left and right parenthesis strings.
String printExpr(
    Expr input,
    Map<int, int> operators,
    String blockLeft,
    String blockRight,
    GetPrintTemplate getTemplate,
    InsertPrintSeparator insertSeparator) {
  // We use a non-recursive approach where we first collect all arguments before
  // retrieving and rendering the template.
  final stack = new List<_PrintStackItem>();

  // Add initial item (after all objects this should be the only thing left in
  // the stack).
  stack.add(new _PrintStackItem(null, null));

  for (var i = 0; i < input.length; i++) {
    final object = input[i];

    // Add object to the stack.
    // Note that objects with 0 arguments could also be added to the argument
    // list of the previous stack. However I decided this is pre-mature
    // optimization.
    stack.add(new _PrintStackItem(object, getTemplate(object)));

    // Try to collapse stack.
    // Note that only the root template is supposed to be null.
    while (stack.last.template != null &&
        stack.last.template.argumentCount == stack.last.arguments.length) {
      // Render template and add to arguments of previous item.
      final item = stack.removeLast();
      final result = printTemplate(item.template, item.parent, item.children,
          item.arguments, operators, blockLeft, blockRight, insertSeparator);
      stack.last.arguments.add(result);
      stack.last.children.add(item.parent);
    }
  }

  // Return generated string.
  return stack.single.arguments.single.result;
}

/// Print the given [template] using the already retrieved [arguments].
PrintResult printTemplate(
    PrintTemplate template,
    ExprObject parent,
    List<ExprObject> children,
    List<PrintResult> arguments,
    Map<int, int> operators,
    String blockLeft,
    String blockRight,
    InsertPrintSeparator insertSeparator) {
  // The buffer must be modifiable, so we use a list.
  final buffer = new List<String>();

  // Some information fields about the resulting string.
  var startsWithNumber = false, endsWithNumber = false;
  var noArgBefore = false, noArgAfter = false;
  var leftFacingPrecedence = -1, rightFacingPrecedence = -1;

  // If the parent is a number, startsWithNumber and endsWithNumber is true.
  if (parent.type == ExprType.number) {
    startsWithNumber = true;
    endsWithNumber = true;
  }

  // Some state data (FSM like). This is not a super elegant solution, perhaps
  // this can be improved later.
  var prevIsTextToken = false;
  var prevArgIndex = 0;
  PrintResult prevArg;

  for (final token in template.tokens) {
    // If the token is not an argument token, add text content.
    if (!token.isArgument) {
      buffer.add(token.text);
      prevIsTextToken = true;
      continue;
    }

    // Throw error if the argument index is invalid.
    if (token.argumentIndex < 0 || token.argumentIndex + 1 > arguments.length) {
      throw new ArgumentError('bad template; invalid argument index');
    }

    // Get argument data.
    final argument = arguments[token.argumentIndex];
    final argObject = children[token.argumentIndex];
    var useParentheses = false;

    if (token.argLeftBaseline) {
      startsWithNumber = startsWithNumber || argument.startsWithNumber;
      noArgBefore = noArgBefore || argument.noArgBefore;
      leftFacingPrecedence = token.blockPrecedenceLevel;
    }
    if (token.argRightBaseline) {
      endsWithNumber = endsWithNumber || argument.endsWithNumber;
      noArgAfter = noArgAfter || argument.noArgAfter;
      rightFacingPrecedence = token.blockPrecedenceLevel;
    }

    // Determine if this argument should be surrounded in parentheses.
    // There are 5 cases when parentheses must be used:
    // 1. The argument is an operator, and has a precedence level lower or
    //    equal to the one indicated by this token, the left or right facing
    //    precedence of the argument is not -1 (indicating there it has an
    //    argument on the baseline), or this template has one argument that is
    //    not on the baseline (e.g. special case `\left(\frac{1}{2}\right)^3`).
    // 2. There is no text token between this parameter and the previous one
    //    and this parameter starts with an integer.
    // 3. There is no text token between this parameter and the previous one
    //    and either one has an noArgBefore/noArgAfter flag set.
    // 4. This template is an operator, and the argument has a left facing
    //    parameter with a parentheses priority lower than this one or
    //    vice versa.
    // 5. This parameter has the noArgBefore flag set and the token has
    //    the paramImposterLeft flag set or vice versa.

    // 1
    if (argObject.type == ExprType.function &&
        operators.containsKey(argObject.value) &&
        operators[argObject.value] <= token.blockPrecedenceLevel &&
        (argument.leftFacingPrecedence != -1 ||
            argument.rightFacingPrecedence != -1 ||
            template.tokens.any((t) =>
                t.isArgument && !t.argLeftBaseline && !t.argRightBaseline))) {
      useParentheses = true;
    }

    // 2
    else if (prevArg != null && !prevIsTextToken && argument.startsWithNumber) {
      useParentheses = true;
    }

    // 3
    else if (prevArg != null &&
        !prevIsTextToken &&
        (prevArg.noArgAfter || argument.noArgBefore)) {
      // By convention: we add parentheses around the parameter that has set
      // the flag. If this is the previous one, we will do some tricks here.
      if (!argument.noArgBefore) {
        buffer.insert(prevArgIndex++, blockLeft);
        buffer.insert(++prevArgIndex, blockRight);
      } else {
        useParentheses = true;
      }
    }

    // 4
    else if (parent.type == ExprType.function &&
        operators.containsKey(parent.value)) {
      final leftCollision = token.argLeftBaseline &&
          argument.rightFacingPrecedence != -1 &&
          argument.rightFacingPrecedence < operators[parent.value];
      final rightCollision = token.argRightBaseline &&
          argument.leftFacingPrecedence != -1 &&
          argument.leftFacingPrecedence < operators[parent.value];

      useParentheses = leftCollision || rightCollision;
    }

    // 5
    else if ((argument.noArgBefore && token.argImposterLeft) ||
        (argument.noArgAfter && token.argImposterRight)) {
      useParentheses = true;
    }

    // Add result to parts.
    if (useParentheses) {
      // Note that since we are adding parentheses, we clear all flags
      // indicating possible collisions.
      prevArgIndex = buffer.length;
      prevArg = new PrintResult(argument.result);

      // Add argument buffer.
      buffer.add('$blockLeft${argument.result}$blockRight');

      // The parenthesis counts as text token.
      prevIsTextToken = true;
    } else {
      // Insert separator to prevent collision.
      if (buffer.isNotEmpty) {
        final separator = insertSeparator(buffer.last, argument.result);
        if (separator.isNotEmpty) {
          buffer.add(separator);
        }
      }

      prevArgIndex = buffer.length;
      prevArg = argument;

      // Add argument to buffer.
      buffer.add(argument.result);

      prevIsTextToken = false;
    }
  }

  // Collect all results into [PrintResult] instance.
  return new PrintResult(
      buffer.join(),
      startsWithNumber,
      endsWithNumber,
      template.noArgBefore || noArgBefore,
      template.noArgAfter || noArgAfter,
      leftFacingPrecedence,
      rightFacingPrecedence);
}

/// Printing stack item.
class _PrintStackItem {
  final ExprObject parent;
  final PrintTemplate template;
  final List<PrintResult> arguments;
  final List<ExprObject> children;

  _PrintStackItem(this.parent, this.template)
      : arguments = [],
        children = [] {
    assert((parent == null && template == null) ||
        parent.argumentCount == template.argumentCount);
  }
}

/// Intermediary render data.
class PrintResult {
  final String result;
  final bool startsWithNumber, endsWithNumber;
  final bool noArgBefore, noArgAfter;

  /// Precedence level at which parentheses are added on the left or right side
  /// of this result. This can be used to detect collisions with encapsulating
  /// templates. For example in the case of `(sin 2a)^2`. The right facing
  /// precedence of the first argument of sin could be the addition operator
  /// (so in this case no parentheses are added), this is smaller than the
  /// precedence level of the power operator.
  final int leftFacingPrecedence, rightFacingPrecedence;

  PrintResult(this.result,
      [this.startsWithNumber = false,
      this.endsWithNumber = false,
      this.noArgBefore = false,
      this.noArgAfter = false,
      this.leftFacingPrecedence = -1,
      this.rightFacingPrecedence = -1]);
}

/// Implementation of [InsertPrintSeparator] for LaTeX strings.
String insertLaTeXSeparator(a, b) {
  final leftAlpha = new RegExp(r'[A-Za-z]$').hasMatch(a);
  final rightAlpha = new RegExp(r'^[A-Za-z]').hasMatch(b);
  return leftAlpha && rightAlpha ? ' ' : '';
}
