// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

library expr.utils;

/// Class for reading a string character by character.
class StringReader {
  /// UTF16 string data
  final List<int> data;

  /// String element pointer
  int ptr = 0;

  StringReader(String str) : data = str.runes.toList();

  /// Check if the reader is at the end of the input string.
  bool get eof => ptr == data.length;

  /// Get current character code.
  int get current => data[ptr];

  /// Test if [current] is equal to the first character in [char].
  bool currentIs(String char) {
    return !eof && current == char.runes.first;
  }

  /// Check if the current character matches the given condition.
  bool checkCurrent(bool condition(int char)) {
    return !eof && condition(current);
  }

  /// Skip all characters that match the given condition.
  void skip(bool condition(int char)) {
    while (checkCurrent(condition)) {
      next();
    }
  }

  /// Skip while character is not the given character.
  void skipUntil(String char) {
    final charIdx = char.runes.first;
    skip((c) => c != charIdx);
  }

  /// Skip all space characters.
  void skipWhitespace() {
    return skip((char) => char == spaceAsciiCode);
  }

  /// Skip all digits.
  void skipDigits() {
    return skip(charIsDigit);
  }

  /// Move to next character.
  int next() {
    assert(!eof);
    return ++ptr;
  }

  /// Move to the next position if the given [condition] it true.
  int nextIf(bool condition) {
    return condition ? next() : ptr;
  }

  /// Check if any of the strings in [values] match at the current pointer
  /// position. If so, move the pointer forward to the first character after the
  /// selected value.
  bool matchAny(Set<String> values) {
    final valuesInts = values.map((v) => v.runes.toList()).toList();
    for (final ints in valuesInts) {
      var offset = 0;
      while (offset < ints.length &&
          ptr + offset < data.length &&
          data[ptr + offset] == ints[offset]) {
        offset++;
      }
      if (offset == ints.length) {
        ptr += offset;
        return true;
      }
    }
    return false;
  }

  /// Get string from [start] to [end] exclusive.
  String str(int start, int end) {
    return new String.fromCharCodes(data.sublist(start, end));
  }
}

const spaceAsciiCode = 32;
const zeroAsciiCode = 48;
const nineAsciiCode = 57;

/// Test if the given [char] is within the given ASCII range.
bool charIsInRange(int char, int start, int endInclusive) {
  return char >= start && char <= endInclusive;
}

/// Test if the given [char] is a decimal digit (0-9).
bool charIsDigit(int char) {
  return charIsInRange(char, zeroAsciiCode, nineAsciiCode);
}
