// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of expr;

/// Print template
///
/// A print template is parsed from a string. The string can contain custom
/// syntax for including arguments. Some LaTeX examples are:
/// - `${.0}+${1(+).}`
/// - `$!-${0(^).}`
/// - `\sin${:0(+).}`
///
/// General syntax:
/// - `$!` at the beginning or end of the string means that no argument, or
///   expression imposter can be added directly before/after this template
///   without first adding parentheses.
/// - `${...}` is an argument insertion point
///
/// Argument syntax:
/// - Argument index
/// - A dot `.` can be used to indicated that the argument is on the baseline
///   (unlike a power for example) and left or right facing in the template
///   after rendering. The side of the dot with respect to the argument index
///   indicates which side its about.
/// - A `(...)` can be used to specify that the argument should be surrounded
///   in parentheses if the argument is an operator with a lower or equal
///   precedence than indicated by the operator character in the parentheses.
/// - A `:` can be used to indicate that the argument is next to an imposter
///   expression. This means that the string expression left/right to this
///   argument may appear as an independent expression, and parentheses should
///   be added around this argument if its template has a `$!`.
class PrintTemplate {
  final List<PrintTemplateToken> tokens;

  /// An argument can never directly be directly before/after this template
  /// (relevant for multiplication).
  final bool noArgBefore, noArgAfter;

  PrintTemplate.fromTokens(this.tokens, this.noArgBefore, this.noArgAfter);

  /// Count number of arguments.
  int get argumentCount =>
      tokens.map((t) => t.argumentIndex).where((i) => i != -1).toSet().length;
}

/// Parse print template from the given [input] string and list of associated
/// [operators].
PrintTemplate parsePrintTemplate(String input, List<Operator> operators) {
  final reader = new StringReader(input);
  final tokens = new List<PrintTemplateToken>();
  var noArgBefore = false, noArgAfter = false;

  var parsedUntil = 0;
  while (!reader.eof) {
    // Skip until next '$'.
    reader.skipUntil(r'$');

    // Index where this argument token starts (if it is one).
    final argTokenStartIdx = reader.ptr;

    // Check there are still characters left.
    if (!reader.eof) {
      // Skip `$`.
      reader.next();

      // Token to be added in the end.
      PrintTemplateToken argumentToken;

      // If the current character is an '!', and this token is in the first or
      // last position in the string, set noArgBefore/After.
      if (reader.currentIs('!')) {
        reader.next();
        if (reader.ptr == 2) {
          noArgBefore = true;
        } else if (reader.eof) {
          noArgAfter = true;
        } else {
          continue;
        }
      } else if (reader.currentIs('{')) {
        reader.next();

        // Parse left indicator.
        final argLeftBaseline = reader.currentIs('.');
        final argImposterLeft = reader.currentIs(':');
        reader.nextIf(argLeftBaseline || argImposterLeft);

        // Try to parse number.
        final argumentIndexStartIdx = reader.ptr;
        reader.skipDigits();
        if (reader.ptr - argumentIndexStartIdx > 0) {
          final idxStr = reader.str(argumentIndexStartIdx, reader.ptr);
          final argumentIndex = int.parse(idxStr);

          // Try to find block precedence level.
          var blockPrecedenceLevel = -1;
          if (reader.currentIs('(')) {
            final startIdx = reader.ptr + 1;
            reader.skipUntil(')');
            final sym = reader.str(startIdx, reader.ptr);
            final ops = operators.where((op) => op.symbols.contains(sym));
            if (ops.isNotEmpty) {
              final op = ops.single;
              blockPrecedenceLevel = op.precedenceLevel;
              reader.next(); // Skip ')'
            } else {
              continue;
            }
          }

          // Parse right indicator.
          final argRightBaseline = reader.currentIs('.');
          final argImposterRight = reader.currentIs(':');
          reader.nextIf(argRightBaseline || argImposterRight);

          // Expect a closing '}' here.
          if (!reader.currentIs('}')) {
            continue;
          }
          reader.next();

          // Create argument token.
          argumentToken = new PrintTemplateToken(
              '',
              argumentIndex,
              blockPrecedenceLevel,
              argLeftBaseline,
              argRightBaseline,
              argImposterLeft,
              argImposterRight);
        } else {
          continue;
        }
      }

      // When reaching this point it means an argument token or $! token was
      // parsed successfully. We need to add the text before this as a raw text
      // token.

      // Add text before this token to tokens.
      if (argTokenStartIdx - parsedUntil > 0) {
        final text = reader.str(parsedUntil, argTokenStartIdx);
        tokens.add(new PrintTemplateToken(text));
      }

      // Move parsing front forward.
      parsedUntil = reader.ptr;

      // If an argument token was produced, add it after the text token.
      if (argumentToken != null) {
        tokens.add(argumentToken);
      }
    }
  }

  // Add final text.
  if (reader.ptr - parsedUntil > 0) {
    tokens.add(new PrintTemplateToken(reader.str(parsedUntil, reader.ptr)));
  }

  return new PrintTemplate.fromTokens(tokens, noArgBefore, noArgAfter);
}

/// Token in a [PrintTemplate]
class PrintTemplateToken {
  /// Raw text
  final String text;

  /// Argument index
  final int argumentIndex;

  /// Block separators (parentheses) must be added when the precedence level of
  /// the argument is smaller or equal to this value.
  final int blockPrecedenceLevel;

  /// After rendering, this argument will be the very first/last thing
  /// (visually) on the initial baseline.
  final bool argLeftBaseline, argRightBaseline;

  /// Left/right to this argument is an imposter expression.
  final bool argImposterLeft, argImposterRight;

  PrintTemplateToken(this.text,
      [this.argumentIndex = -1,
      this.blockPrecedenceLevel = -1,
      this.argLeftBaseline = false,
      this.argRightBaseline = false,
      this.argImposterLeft = false,
      this.argImposterRight = false]);

  /// Test if this is an argument token.
  bool get isArgument {
    assert(text.isNotEmpty || argumentIndex != -1);
    return argumentIndex != -1;
  }
}

/// Determine if a given [template] has exactly one argument token for each
/// argument index from 0 to [PrintTemplate.argumentCount].
bool checkPrintTemplateArguments(PrintTemplate template) {
  final argumentIndices =
      template.tokens.map((t) => t.argumentIndex).where((i) => i != -1).toSet();
  return argumentIndices.isEmpty ||
      (min(argumentIndices) == 0 &&
          max(argumentIndices) + 1 == argumentIndices.length);
}
