// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of expr;

/// Generate true statement for a given transform.
typedef TrueStatement<M> GenerateTrueStatement<M>(Transform t);

/// Find a list of detailed rewrites to prove the given [statement]. Searches
/// through [formulas] for pre-defined rules. If the algorithm gets stuck
/// (cannot resolve a step), it will return null. The recursive premise
/// resolving depth can be limited by setting [recursionDepth].
///
/// [getExpansion]: Generate formula for a numerical expansion. This means that
/// the transform target can be evaluated to form the transform source. To apply
/// this in a forward manner a formula is required. This function should
/// generate such a formula.
///
/// [getAssumption]: Generate formula for a transform that cannot be proven by
/// any of the implemented means. Usually this is an assumption. If the returned
/// statement is null, this function terminates by returning null.
///
/// Note: Currently premises are only resolved in a backward manner (first the
/// proposition has to match, then the premises are checked). In some scenarios,
/// it is useful to find premises in a forward manner (match the proposition
/// only partially and then search for premise proofs to eliminate the rest of
/// the mapping).
Lemma<M> proveLemma<M>(
    Transform statement,
    Set<int> canEval,
    EvalExpr eval,
    Set<int> canRearrange,
    Map<DistFnKey, List<int>> canRedistribute,
    Iterable<Formula<M>> formulas,
    GenerateTrueStatement<M> getExpansion,
    GenerateTrueStatement<M> getAssumption,
    {int recursionDepth: -1}) {
  final rewrites = new List<Rewrite<M>>();
  final queue = new List<Transform>();
  queue.add(statement);

  QUEUE:
  while (queue.isNotEmpty) {
    final t = queue.removeLast();
    if (isIdentityTransform(t)) {
      continue QUEUE;
    } else {
      // Try evaluating.
      final evaluate = new Set<int>();
      final evalResult = tryEvaluateExpr(t.source, canEval, eval, evaluate);
      if (evalResult == t.target) {
        rewrites.addAll(getEvaluateRewrites<M>(evaluate, t.source.offset));
        continue QUEUE;
      }

      // Try rearranging.
      final rlist = findRearrangements(t, canRearrange);
      if (rlist.isNotEmpty) {
        // Sorting is not necessary since rearranged sub-expressions retain the
        // same length. Nested rearrangements are first in the list.
        rewrites.addAll(rlist.map((r) {
          return new Rewrite<M>(
              r.index + t.source.offset, RewriteInstruction.rearrange,
              parameters: r.format);
        }));
        continue QUEUE;
      }

      // Try redistributing.
      final redist = findRedistribution(t, canRearrange, canRedistribute);
      if (redist != null) {
        rewrites.add(new Rewrite<M>(
            t.source.offset, RewriteInstruction.redistribute,
            parameters: encodeRedistribution(redist)));
        continue QUEUE;
      }

      // Try a numerical expansion.
      if (t.source.first.type == ExprType.number) {
        if (tryEvaluateExpr(t.target, canEval, eval) == t.source) {
          final statement = getExpansion(t);
          assert(statement != null);
          rewrites.add(new Rewrite<M>(
              t.source.offset, RewriteInstruction.substitute,
              statement: statement));
          continue QUEUE;
        }
      }

      // Search for formula.
      FORMULAS:
      for (final formula in formulas) {
        // Skip if [recursionDepth] is zero and this formula has premises.
        if (recursionDepth == 0 && formula.premises.isNotEmpty) {
          continue FORMULAS;
        }

        // Parameters for [compareTransforms].
        final mapping = new Mapping();
        final eval1 = new Set<int>(), eval2 = new Set<int>();

        // Test if the formula proposition matches.
        if (compareTransforms(t, formula.proposition, canEval, eval,
            mapping: mapping, evaluateBefore: eval1, evaluateAfter: eval2)) {
          // If there are premises, we must try to prove each via a lemma.
          final premiseProofs = new List<Lemma>();
          for (final premise in formula.premises) {
            final l = mapTransform(premise, mapping);
            final proof = proveLemma(l, canEval, eval, canRearrange,
                canRedistribute, formulas, getExpansion, getAssumption,
                recursionDepth: recursionDepth - 1);

            // If no proof is found, discard this formula.
            if (proof == null) {
              continue FORMULAS;
            } else {
              premiseProofs.add(proof);
            }
          }

          // Add pre evaluation rewrites.
          rewrites.addAll(getEvaluateRewrites<M>(eval1, t.source.offset));

          // Add substitution.
          final trueStatement = new TrueStatement<M>(formula, premiseProofs);
          rewrites.add(new Rewrite<M>(
              t.source.offset, RewriteInstruction.substitute,
              statement: trueStatement));

          // Add post evaluation rewrites.
          rewrites.addAll(getEvaluateRewrites<M>(eval2, t.source.offset));

          continue QUEUE;
        }
      }

      // Split into argument transforms.
      final arguments = getArgumentTransforms(t);
      if (arguments.isNotEmpty) {
        // Resolve per argument by adding each one to the queue.
        // Note that we go backward (from the last to the first argument) in
        // order to preserve the object indexing.
        queue.addAll(arguments);
        continue QUEUE;
      } else {
        // Not possible to split into arguments: try processing as assumption.
        final assumption = getAssumption(t);
        if (assumption != null) {
          rewrites.add(new Rewrite<M>(
              t.source.offset, RewriteInstruction.substitute,
              statement: assumption));
        } else {
          return null;
        }
      }
    }
  }

  final lemma = new Lemma.from(statement.source, rewrites, eval);
  assert(lemma.statement.target == statement.target);
  return lemma;
}
