// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of expr;

/// Get mapping of premise proofs in the given true [statement].
Mapping computeStatementMapping(TrueStatement statement) {
  final mapping = new Mapping();
  for (var i = 0; i < statement.formula.premises.length; i++) {
    final premise = statement.formula.premises[i];
    final proof = statement.premiseProofs[i];

    // Find mapping of proof and mapping between premise and proof.
    final t1 = new Transform(premise.source, proof.statement.source);
    final t2 = new Transform(premise.target, proof.statement.target);
    mergeMappings(mapping, findSmallestTransforms(t1));
    mergeMappings(mapping, findSmallestTransforms(t2));
  }
  return mapping;
}

/// Apply [rewrite] to [input].
Expr applyRewrite(Expr input, Rewrite rewrite, EvalExpr eval) {
  switch (rewrite.instruction) {
    case RewriteInstruction.evaluate:
      return applyEvaluation(input, rewrite.position, eval);

    case RewriteInstruction.rearrange:
      final r = new Rearrangement(rewrite.position, rewrite.parameters);
      return rearrangeExpr(input, r);

    case RewriteInstruction.redistribute:
      // Note that [decodeRedistribution] empties the provided list.
      final r = decodeRedistribution(new List<int>.from(rewrite.parameters));
      final subExpr = getSubExpr(input, rewrite.position);
      final replacement = redistributeExpr(subExpr, r);
      return replaceSubExpr(input, rewrite.position, replacement);

    case RewriteInstruction.substitute:
      final mapping = computeStatementMapping(rewrite.statement);
      if (!isNormalizedMapping(mapping)) {
        throw new ArgumentError('statement has unnormalized proof');
      }

      final proposition = rewrite.statement.formula.proposition;
      return substituteTransform(input, rewrite.position, proposition, mapping);

    default:
      throw new ArgumentError('unable to process rewrite instruction');
  }
}

/// Apply all [rewrites] to [source].
Expr applyRewrites(Expr source, List<Rewrite> rewrites, EvalExpr eval) {
  return rewrites.fold(source, (expr, r) => applyRewrite(expr, r, eval));
}

/// Convert set of evaluation [indices] to list of rewrites.
List<Rewrite<M>> getEvaluateRewrites<M>(Set<int> indices, int offset) {
  // Evaluate from big to small index to prevent index shifting.
  final sorted = indices.toList()..sort((a, b) => b - a);
  final out = sorted.map((index) {
    return new Rewrite<M>(index + offset, RewriteInstruction.evaluate);
  });
  return out.toList();
}
