// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of expr;

enum CollapseMode {
  /// Do not combine with another rewrite with mode [standalone] or [adjacent].
  standalone,

  /// Do not combine with an overlapping rewrite with mode [adjacent].
  adjacent,

  /// Add after [standalone] or [adjacent] rewrite in the same part. Move to the
  /// next part when a [prepend] rewrite has been encountered.
  append,

  /// Add after [standalone] or [adjacent] rewrite in the next part.
  prepend
}

/// Split the [input] lemma into multiple smaller consecutive lemmas according
/// to the given settings.
List<Lemma> splitLemma(
    Lemma input, EvalExpr eval, Set<DistFnKey> canRedistribute,
    {CollapseMode evaluate(Expr target),
    CollapseMode rearrange(int f),
    CollapseMode redistribute(DistFnKey key),
    CollapseMode substitute(TrueStatement statement)}) {
  final _evaluate = evaluate ?? (_) => CollapseMode.append;
  final _rearrange = rearrange ?? (_) => CollapseMode.append;
  final _redistribute = redistribute ?? (_) => CollapseMode.adjacent;
  final _substitute = substitute ?? (_) => CollapseMode.adjacent;

  // Split into multiple rewrite sequences.
  final parts = new List<_LemmaPart>();
  var expr = input.statement.source;
  for (final rewrite in input.rewrites) {
    final subExpr = getSubExpr(expr, rewrite.position);
    final coord = _computeExprIndexCoord(expr, rewrite.position);
    expr = rewrite.result ?? applyRewrite(expr, rewrite, eval);

    // Find collapse mode for rewrite.
    var mode = CollapseMode.append;
    switch (rewrite.instruction) {
      case RewriteInstruction.evaluate:
        mode = _evaluate(subExpr);
        break;

      case RewriteInstruction.rearrange:
        mode = _rearrange(subExpr.first.value);
        break;

      case RewriteInstruction.redistribute:
        final key = _findDistFnKey(subExpr, canRedistribute);
        mode = _redistribute(key);
        break;

      case RewriteInstruction.substitute:
        mode = _substitute(rewrite.statement);
        break;
    }

    // Add rewrite to parts.
    final _rewrite = new _LemmaRewrite(rewrite, coord);
    switch (mode) {
      case CollapseMode.standalone:
        // Add new part for this rewrite and mark it for [adjacent] exclusion.
        parts.add(new _LemmaPart(allowAdjacent: false));
        parts.last.rewrites.add(_rewrite);
        parts.last.occupation.add(coord);
        break;

      case CollapseMode.adjacent:
        // Find first part this rewrite has no overlap with.
        final part = _getFirstNonOverlap(parts, coord);
        part.rewrites.add(_rewrite);
        part.occupation.add(coord);
        break;

      case CollapseMode.append:
        // Find last part this rewrite has overlap with. This includes prepends.
        final overlap = parts.where((p) =>
            _hasOverlap(p.occupation, coord) ||
            _hasOverlap(p.occupationPre, coord));
        final part = overlap.isEmpty
            ? (parts.isEmpty ? (parts..add(new _LemmaPart())).last : parts[0])
            : overlap.last;
        part.rewrites.add(_rewrite);
        break;

      case CollapseMode.prepend:
        // Find first part this rewrite has no overlap with.
        final part = _getFirstNonOverlap(parts, coord);
        part.rewrites.add(_rewrite);
        // Add to prepend occupation buffer to redirect overlapping appends on
        // the previous part to this part.
        part.occupationPre.add(coord);
        break;
    }
  }

  // Build output list.
  final lemmas = new List<Lemma>();
  expr = input.statement.source;
  for (final part in parts) {
    final source = new Expr.copy(expr);
    final rewrites = new List<Rewrite>();
    for (final r in part.rewrites) {
      final position = _computeExprCoordIndex(expr, r.coord);
      final rewrite = new Rewrite(position, r.rewrite.instruction,
          parameters: r.rewrite.parameters, statement: r.rewrite.statement);
      expr = applyRewrite(expr, rewrite, eval);
    }
    lemmas.add(new Lemma(new Transform(source, expr), rewrites));
  }
  assert(expr == input.statement.target);
  return lemmas;
}

/// Compute the argument index list for the given [index] in the given [expr].
List<int> _computeExprIndexCoord(Expr expr, int index) {
  final coord = new List<List<int>>();
  for (var i = 0; i < index; i++) {
    final obj = expr[i];
    coord.add([0, obj.argumentCount]);
    while (coord.last[0] == coord.last[1]) {
      coord.removeLast();
      coord.last[0]++;
    }
  }
  return coord.map((c) => c[0]).toList();
}

/// Compute the index in [expr] for the argument index list in [coord].
int _computeExprCoordIndex(Expr expr, List<int> coord) {
  var i = 0;
  for (final argi in coord) {
    i += 1 + skipObjectContent(expr, i + 1, argi);
  }
  return i;
}

/// Determine if [coord] overlaps with [occupied].
bool _hasOverlap(List<List<int>> occupied, List<int> coord) {
  return occupied.any((oc) {
    // Check if both lists share the same root.
    for (var i = 0; i < coord.length && i < oc.length; i++) {
      if (coord[i] != oc[i]) {
        return false; // Not the same root (overlap)
      }
    }
    // Same root
    return true;
  });
}

/// Return the first part in [parts] that has no overlap with [coord]. A new
/// part may need to be added.
_LemmaPart _getFirstNonOverlap(List<_LemmaPart> parts, List<int> coord) {
  final noOver = parts.where((p) => !_hasOverlap(p.occupation, coord));
  return noOver.isEmpty ? (parts..add(new _LemmaPart())).last : noOver.first;
}

class _LemmaRewrite {
  final Rewrite rewrite;
  final List<int> coord;
  _LemmaRewrite(this.rewrite, this.coord);
}

class _LemmaPart {
  final bool allowAdjacent;
  final rewrites = new List<_LemmaRewrite>();

  /// Coordinates that are affected by a [CollapseMode.prepend] rewrite.
  final occupationPre = new List<List<int>>();

  /// Coordinates are affected by a [CollapseMode.standalone] or
  /// [CollapseMode.adjacent] rewrite.
  final occupation = new List<List<int>>();

  _LemmaPart({this.allowAdjacent: true});
}
