// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of expr;

/// Mathematical formula where [proposition] is true if the [premises] are true.
/// In other words; [proposition] can be applied if the [premises] are proven.
/// A formula is a statement which proposition is true when all premises can be
/// shown to be true.
class Formula<M> {
  final int id;
  final List<Transform> premises;
  final Transform proposition;
  final M metadata;

  Formula(this.premises, this.proposition, {this.id, this.metadata});
}

/// Rewrite instructions
enum RewriteInstruction { evaluate, rearrange, redistribute, substitute }

/// Note: A rearrangement is a special rewrite that can be applied when some
/// properties of a function are known (if it is both commutative and
/// associative such as addition). This case is built in to simplify rewrite
/// sequences.
///
/// Set [parameters] for [RewriteInstruction.rearrange]
/// Set [statement] for [RewriteInstruction.substitute]
class Rewrite<M> {
  final int position;
  final RewriteInstruction instruction;
  final List<int> parameters;
  final TrueStatement<M> statement;

  /// Metadata for compatibility with the QEDb schema. By convention there is no
  /// obligation to retain this value during processing.
  Expr result;

  Rewrite(this.position, this.instruction,
      {this.parameters, this.statement, this.result}) {
    assert(instruction != RewriteInstruction.rearrange || parameters != null);
    assert(instruction != RewriteInstruction.substitute || statement != null);
  }
}

/// A formula which proposition is true because all premises are validated.
class TrueStatement<M> {
  final Formula<M> formula;
  final List<Lemma<M>> premiseProofs;

  TrueStatement(this.formula, this.premiseProofs);
}

/// A sequence of rewrites is called a lemma following the general slang often
/// used in this field. When a lemma is constructed the statement is computed.
class Lemma<M> {
  final int id;
  final Transform statement;
  final List<Rewrite<M>> rewrites;

  Lemma(this.statement, this.rewrites, {this.id}) {
    assert(statement.source.offset == statement.target.offset);
  }

  factory Lemma.from(Expr source, List<Rewrite<M>> rewrites, EvalExpr eval) {
    final target = applyRewrites(source, rewrites, eval);
    final statement = new Transform(source, target);
    return new Lemma(statement, rewrites);
  }
}

/// Convert a [Redistribution] into an integer array.
List<int> encodeRedistribution(Redistribution r) {
  final list = new List<int>();
  void writeNode(RedistrNode n) {
    list.add(n.sourceIndex);
    if (n.sourceIndex == -1) {
      writeNode(n.a);
      writeNode(n.b);
      list.add(n.mergePoint.length);
      list.addAll(n.mergePoint);
    } else {
      if (n.r != null) {
        list..add(n.r.format.length)..add(n.r.index);
        list.addAll(n.r.format);
      } else {
        list.add(0);
      }
    }
  }

  list..add(r.key.f)..add(r.key.g)..add(r.fGapArgs.length);
  list.addAll(r.fGapArgs);
  writeNode(r.node);
  return list;
}

/// Convert an integer array into a [Redistribution].
Redistribution decodeRedistribution(List<int> input) {
  int get1() => input.removeAt(0);
  List<int> get(int length) => new List<int>.generate(length, (_) => get1());

  RedistrNode readNode() {
    final sourceIndex = get1();
    if (sourceIndex == -1) {
      final a = readNode(), b = readNode();
      return new RedistrNode(sourceIndex, null, a, b, get(get1()));
    } else {
      final flen = get1();
      final r = flen > 0 ? new Rearrangement(get1(), get(flen)) : null;
      return new RedistrNode(sourceIndex, r, null, null, null);
    }
  }

  final key = new DistFnKey(get1(), get1());
  final fGapArgs = get(get1());
  final rootNode = readNode();
  return new Redistribution(key, fGapArgs, rootNode);
}
