// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of expr;

// The algorithms described here are ad-hoc implementations because there is
// currently not a more general use case.

/// Check if the given [transform] is a rearrangement pattern. Returns a set
/// containing the ID of rearrangeable function (empty otherwise).
Set<int> detectRearrangementPattern(Transform transform) {
  final s = transform.source, t = transform.target;
  // f(#x, #y) => f(#y, #x)
  final set = new Set<int>();
  if (s.length == 3 && t.length == 3) {
    final f1 = s[0], x1 = s[1], y1 = s[2];
    final f2 = t[0], y2 = t[1], x2 = t[2];

    if ((f1 == f2 && x1 == x2 && y1 == y2) &&
        (_fn(f1, 2) && _pat(x1, 0) && _pat(y1, 0))) {
      set.add(f1.value);
    }
  }
  return set;
}

/// Check if the given [transform] is a redistribution pattern. Returns IDs of
/// redistributable functions or an empty map if it is not such a pattern.
/// Note that an error is thrown if the transform data is not complete.
Map<DistFnKey, List<int>> detectRedistributionPattern(Transform transform) {
  // f(g(#x, #y)) => g(f(#x), f(#y)) where f may have more arguments.

  // Get ExprObject for f and g, and check if they meet the expectation.
  final s = transform.source, t = transform.target;
  final f = s.first, g = t.first;
  if (!(f.type == ExprType.function && _fn(g, 2))) {
    return {};
  }

  // Extract more data about f and g.
  final gTargetArgs = getExprArguments(t);
  final fSourceArgs = getExprArguments(s);
  final fTargetArgs1 = getExprArguments(gTargetArgs[0]);
  final fTargetArgs2 = getExprArguments(gTargetArgs[1]);

  // Check if the arguments of g in the target are both the f function.
  if (!(gTargetArgs[0].first == f && gTargetArgs[1].first == f)) {
    return {};
  }

  // Find argument index of g in source.
  final gArgi = indexOf(fSourceArgs, (arg) => arg.first == g);
  if (gArgi == -1) {
    return {};
  }

  // Check if all other arguments of f except [gArgi] are the same pattern
  // symbols (meaning all other arguments can be copied over when substituted).
  final fArgs = [fSourceArgs, fTargetArgs1, fTargetArgs2];
  for (var i = 0; i < fArgs.first.length; i++) {
    if (i != gArgi) {
      // Argument i must be a pattern symbol and the arguments in the other
      // argument lists at i should be equal.
      final argIsPat = _pat(fArgs.first[i].first, 0);
      final sameArgs = fArgs.every((args) => args[i] == fArgs.first[i]);
      if (!argIsPat || !sameArgs) {
        return {};
      }
    }
  }

  // Get x and y and check if they are both pattern symbols.
  final gSourceArgs = getExprArguments(fSourceArgs[gArgi]);
  final x = gSourceArgs[0], y = gSourceArgs[1];
  if (!(_pat(x.first, 0) && _pat(y.first, 0))) {
    return {};
  }

  // Check if x and y are the same as the arguments at [gArgi] in the target.
  final xTarget = fTargetArgs1[gArgi], yTarget = fTargetArgs2[gArgi];
  if (x != xTarget || y != yTarget) {
    return {};
  }

  // Return redistribution settings.
  final out = new Map<DistFnKey, List<int>>();
  out[new DistFnKey(f.value, g.value)] = [gArgi];
  return out;
}

// helpers
bool _fn(ExprObject obj, int argc) =>
    obj.type == ExprType.function && obj.argumentCount == argc;
bool _pat(ExprObject obj, int argc) =>
    obj.type == ExprType.pattern && obj.argumentCount == argc;
