// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

library expr.test.testenv;

import 'package:expr/expr.dart';
import 'package:quiver/core.dart';

// Rainbow tables for [getTemplate] that are populated by [getId].
final _rainbowTableFunctions = new Map<int, String>();
final _rainbowTablePatterns = new Map<int, String>();

/// [GetExprObjectId] implementation
int getId(ExprType type, String label) {
  final id = hash2(type, label);

  if (type == ExprType.function) {
    _rainbowTableFunctions[id] = label;
  } else if (type == ExprType.pattern) {
    _rainbowTablePatterns[id] = label;
  }

  return id;
}

/// Shortcut to get a function object ID.
int fnId(String label) => getId(ExprType.function, label);

/// Available templates for function IDs.
final functionTemplates = new Map<int, PrintTemplate>();

/// [GetPrintTemplate] implementation
PrintTemplate getTemplate(ExprObject object, String argsLeft, String argsRight,
    String argsSeparator, String fnLabel(String input)) {
  switch (object.type) {
    case ExprType.number:
      final prefix = object.value.isNegative ? '\$!' : '';
      return parsePrintTemplate('$prefix${object.value}', operators);

    case ExprType.function:
    case ExprType.pattern:
      final isFunction = object.type == ExprType.function;
      if (isFunction && functionTemplates.containsKey(object.value)) {
        return functionTemplates[object.value];
      }

      // Generate template from label.
      final label = isFunction
          ? _rainbowTableFunctions[object.value]
          : _rainbowTablePatterns[object.value];

      var template = label;
      if (object.type == ExprType.pattern) {
        template = '#$template';
      }

      // Add argument tokens.
      if (object.argumentCount == 0) {
        if (label.length > 1) {
          template = '\$!$template\$!';
        }
      } else {
        final buffer = new StringBuffer(fnLabel(template));
        buffer.write(argsLeft);
        for (var i = 0; i < object.argumentCount; i++) {
          buffer.write('\${$i}');
          if (i != object.argumentCount - 1) {
            buffer.write(argsSeparator);
          }
        }
        buffer.write(argsRight);
        template = buffer.toString();
      }

      return parsePrintTemplate(template, operators);

    default:
      throw new UnimplementedError('unknown object type');
  }
}

/// Generate list of operators for testing.
List<Operator> getTestOperators(int imid) {
  final ops = new List<Operator>();
  void addOperator(
      int lvl, List<String> lbls, Associativity a, OperatorNotation notation) {
    final functionId = fnId(lbls.first);
    ops.add(new Operator(functionId, lvl, lbls, notation, a));
  }

  // Add all tests operators.
  addOperator(1, ['=>', '-->'], Associativity.ltr, OperatorNotation.infix);
  addOperator(1, ['='], Associativity.ltr, OperatorNotation.infix);
  addOperator(2, ['+'], Associativity.ltr, OperatorNotation.infix);
  addOperator(2, ['-'], Associativity.ltr, OperatorNotation.infix);
  addOperator(3, ['*'], Associativity.ltr, OperatorNotation.infix);
  addOperator(3, ['/'], Associativity.ltr, OperatorNotation.infix);
  addOperator(4, ['^'], Associativity.rtl, OperatorNotation.infix);
  addOperator(5, ['~', '-'], Associativity.rtl, OperatorNotation.prefix);
  addOperator(6, ['!'], Associativity.ltr, OperatorNotation.postfix);

  // Implicit multiplication is set to right associativity by default so that
  // expressions like these can be written: `a^2b` (which would otherwise be
  // parsed as `(a^2)*b`). Increasing the precedence level is not an option,
  // this would result in: `2a^b` => `(2*a)^b`.
  ops.add(new Operator(imid, 4, [], OperatorNotation.infix, Associativity.rtl));

  return ops;
}

/// Load the given function [templates].
void loadTemplates(Map<String, String> templates) {
  templates.forEach((label, templateSource) {
    final template = parsePrintTemplate(templateSource, operators);
    functionTemplates[fnId(label)] = template;
  });
}

/// Reset templates
void clearTemplates() {
  functionTemplates.clear();
}

/// Some default print templates.
final defaultTemplates = {
  '=': r'${.0} = ${1(+).}',
  '+': r'${.0} + ${1(+).}',
  '-': r'${.0} - ${1(+).}',
  '*': r'${.0(+):}*${:1(*).}',
  '/': r'${.0(+):}/${:1(*).}',
  '^': r'${.0(*)}^${1(+)}',
  '~': r'$!¬${0(^).}',
};

/// Print [input] expression.
String easyPrintExpr(Expr input) {
  final op2Pre = new Map<int, int>.fromIterable(operators,
      key: (op) => op.functionId, value: (op) => op.precedenceLevel);
  return printExpr(input, op2Pre, '(', ')', (obj) {
    return getTemplate(obj, '(', ')', ', ', (str) => str);
  }, (l, r) => '');
}

/// All available evaluation computations.
final computations = {
  fnId('+'): (List<num> a) => a[0] + a[1],
  fnId('-'): (List<num> a) => a[0] - a[1],
  fnId('*'): (List<num> a) => a[0] * a[1],
  fnId('~'): (List<num> a) => -a[0]
};

/// Implementation of [EvalExpr]
num eval(int functionId, List<num> arguments) {
  return computations[functionId](arguments);
}

// Add some useful variables to the global scope (remember this is just for
// testing).
final imid = 0;
final mid = fnId('*');
final operators = getTestOperators(imid);
final Set<int> canEval = computations.keys.toSet();
final Set<int> canRearrange = ['+', '*'].map(fnId).toSet();
final Map<DistFnKey, List<int>> canRedistribute = {
  new DistFnKey(fnId('*'), fnId('+')): [0, 1]
};

// Parsing helper.
Expr parse(String input) {
  return parseExpr(input, operators, getId, '#', imid, mid);
}

// Shorthand for parsing a transform.
Transform parseTransform(String input) {
  final expr = parse(input);

  // We assume the top level object is a transform function (=>). We simply
  // split the expression into arguments.
  final from = getSubExpr(expr, 1);
  final to = getSubExpr(expr, 1 + from.length);

  // Create new instances to reset offset.
  return new Transform(new Expr(from), new Expr(to));
}

// Use for building expressions without the parser.
typedef Expr ExprBuilder([Expr a1, Expr a2, Expr a3, Expr a4]);

Expr n(num value) {
  final type = ExprType.number;
  return new Expr([new ExprObject(type, value, 0)]);
}

Expr symbol(String label) {
  final type = ExprType.function;
  return new Expr([new ExprObject(type, getId(type, label), 0)]);
}

Expr buildExprObject(ExprType type, String label, List<Expr> args) {
  args.removeWhere((argument) => argument == null);
  final id = getId(type, label);
  final exprContent = new List<ExprObject>();
  exprContent.add(new ExprObject(type, id, args.length));
  args.forEach(exprContent.addAll);
  return new Expr(exprContent);
}

ExprBuilder function(String label) {
  return ([a1, a2, a3, a4]) =>
      buildExprObject(ExprType.function, label, [a1, a2, a3, a4]);
}

ExprBuilder pattern(String label) {
  return ([a1, a2, a3, a4]) =>
      buildExprObject(ExprType.pattern, label, [a1, a2, a3, a4]);
}

/// Implementation of proveLemma.getExpansion.
TrueStatement getExpansion(Transform t) {
  return new TrueStatement(new Formula([], t), []);
}

/// Implementation of proveLemma.getAssumption.
TrueStatement getAssumption(Transform t) => null;
