// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

library expr.test.core_test;

import 'package:test/test.dart';
import 'package:expr/expr.dart';
import 'package:expr/testenv.dart';

void main() {
  test('compareExprObjects', () {
    final expr = parse('a + a');
    expect(compareExprObjects(expr[1], expr[2]), isTrue);
    expect(expr[1] == expr[2], isTrue);
  });

  test('Transform.hashCode', () {
    final formula = parseTransform('f(#x) => #x^2');
    final f1 = parse('f(1)');
    final t1 = parseTransform('f(1) => 1^2');
    final t2 = new Transform(f1, substituteTransform(f1, 0, formula));

    expect(t1 == t2, isTrue);
    expect(t1.hashCode, equals(t2.hashCode));
  });

  test('evaluateExpr', () {
    expect(evaluateExpr(parse('99.9'), eval), equals(99.9));
    expect(evaluateExpr(parse('1 + (2 - (3 * 4))'), eval), equals(-9));
  });

  test('tryEvaluateTransform', () {
    final t1 = parseTransform('a + 4 => a + (7 - 3)');
    final t2 = tryEvaluateTransform(t1, canEval, eval);
    expect(isIdentityTransform(t2), isTrue);
  });

  test('invertTransforms', () {
    final t1 = parseTransform('a + b => c - d');
    final t2 = parseTransform('c - d => a + b');
    final it = invertTransforms([t1, t2]);
    expect(it[0].source, equals(t1.target));
    expect(it[0].target, equals(t1.source));
    expect(it[1].source, equals(t2.target));
    expect(it[1].target, equals(t2.source));
  });

  test('findSmallestTransforms', () {
    final t = parseTransform('a + b => 1 + (c + 2)');
    final st = findSmallestTransforms(t);
    final expected = {
      parse('a'): [parse('1')].toSet(),
      parse('b'): [parse('c + 2')].toSet()
    };

    expect(st, equals(expected));
  });

  test('findRearrangements', () {
    // Find rearrangements.
    final t3f = 'a * b * (a + (d * (e * f)) + c)';
    final t3t = '((f * e * d) + c + a) * (b * a)';
    final t3 = parseTransform('$t3f => $t3t');
    final r = findRearrangements(t3, canRearrange);
    expect(r.length, equals(3));
    expect(r[0].index, equals(7));
    expect(r[0].format, equals([-1, -1, 2, 1, 0]));
    expect(r[1].index, equals(4));
    expect(r[1].format, equals([-1, -1, 1, 2, 0]));
    expect(r[2].index, equals(0));
    expect(r[2].format, equals([-1, 2, -1, 1, 0]));

    // Validate by application.
    final rres = r.fold(t3.source, rearrangeExpr);
    expect(rres, equals(t3.target));

    // Test non-rearrangeable cases.
    final t4 = parseTransform('a / b => b / a');
    expect(findRearrangements(t4, canRearrange).isEmpty, isTrue);
    final t5 = parseTransform('a * a => a * b');
    expect(findRearrangements(t5, canRearrange).isEmpty, isTrue);
  });

  test('findRedistributions', () {
    final key = new DistFnKey(fnId('*'), fnId('+'));

    // Test unwinding.
    final expr = parse('(a + b)*(c + d)*(e + f)');
    final unwinded = flattenMergeTree(unwindExprDist(expr, key, [0, 1]));

    expect(
        unwinded,
        equals([
          '(a c) e',
          '(b c) e',
          '(a d) e',
          '(b d) e',
          '(a c) f',
          '(b c) f',
          '(a d) f',
          '(b d) f',
        ].map(parse).toList()));

    // Test redistribution finding and application.
    final t1 = parseTransform('(a + b)(c + d) => c*(a + b) + d*a + b*d');
    final r1 = findRedistribution(t1, canRearrange, canRedistribute);
    final t2 = invertTransform(t1);
    final r2 = findRedistribution(t2, canRearrange, canRedistribute);

    expect(r1, isNotNull);
    expect(r2, isNotNull);
    expect(redistributeExpr(parse('(a + b)(c + d)'), r1), equals(t1.target));
    expect(redistributeExpr(t2.source, r2), equals(parse('(a + b)(c + d)')));

    // Check that algorithm is terminated if g function is not rearrangeable.
    final _canRearrange = canRearrange.toSet()..remove(key.g);
    expect(findRedistribution(t1, _canRearrange, canRedistribute), isNull);
  });

  test('compareTransforms', () {
    // Test for simple pattern.
    final t1 = parseTransform('diff(f, f^5) => 5f^4');
    final t2 = parseTransform('diff(#x, #x^#n) => #n#x^(#n - 1)');
    expect(compareTransforms(t1, t2, canEval, eval), isTrue);

    // Test for patterns with arguments.
    final t3 = parseTransform('x*(a(x)b(x)) => x*(a(x + 2)b(x + 2))');
    final t4 = parseTransform('#x#f(#x) => #x#f(#x + 2)');
    expect(compareTransforms(t3, t4, canEval, eval), isTrue);

    // Test nested pattern.
    final t5 = parseTransform('#a*(#b + #c) => #a#b + #a#c');
    final t6 = parseTransform('#f(#g(#x, #y)) => #g(#f(#x), #f(#y))');

    // This doesn't work without prior mapping for #g, #x and #y.
    final m = new Mapping();
    m[parse('#g(#x, #y)')] = [parse('#b + #c')].toSet();
    m[parse('#x')] = [parse('#b')].toSet();
    m[parse('#y')] = [parse('#c')].toSet();

    expect(compareTransforms(t5, t6, canEval, eval), isFalse);
    expect(compareTransforms(t5, t6, canEval, eval, mapping: m), isTrue);
  });

  test('compareTransforms with conflicts', () {
    final evltd = new Set<int>();

    // Test with existing conflict.
    final t1 = parseTransform('(2 + 3)(10 - 4)(2 * 3) => 6^3');
    final t2 = parseTransform('#a#a#a => #a^3');
    expect(compareTransforms(t1, t2, canEval, eval), isFalse);

    // Test with conflict that can be resolved through evaluation.
    // Note that implicit multiplication is right associative!
    evltd.clear();
    final t3 = parseTransform('(2 + 4)(10 - 4)(2 * 3) => 6^3');
    final t4 = parseTransform('#a#a#a => #a^3');
    final r1 = compareTransforms(t3, t4, canEval, eval, evaluateBefore: evltd);
    expect(r1, isTrue);
    expect(evltd, equals([1, 5, 8].toSet()));

    // Test with evaluation afterwards.
    evltd.clear();
    final t5 = parseTransform('(2 + 4)(10 - 4) => 30 + 6');
    final t6 = parseTransform('#a#a => (#a - 1)#a + #a');
    final r2 = compareTransforms(t5, t6, canEval, eval, evaluateAfter: evltd);
    expect(r2, isTrue);
    expect(evltd, equals([1].toSet()));
  });

  test('getArgumentTransforms', () {
    // Check that t1 and t2 are not comparable.
    final t1 = parseTransform('5 + diff(f, f^5) => 2 + 3 + 5f^4');
    final t2 = parseTransform('diff(#x, #x^#n) => #n#x^(#n - 1)');
    expect(compareTransforms(t1, t2, canEval, eval), isFalse);

    // Get argument transforms and check that the first one can be evaluated and
    // the second one matches pattern t2.
    final a = getArgumentTransforms(t1);
    expect(a.length, equals(2));

    final a0eval = tryEvaluateTransform(a[0], canEval, eval);
    expect(isIdentityTransform(a0eval), isTrue);
    expect(compareTransforms(a[1], t2, canEval, eval), isTrue);

    // Check that the argument transforms of two numbers or incompatible
    // patterns/functions is empty.
    final t3 = parseTransform('1 => 2');
    final t4 = parseTransform('#f(1) => #g(1)');
    expect(getArgumentTransforms(t3).isEmpty, isTrue);
    expect(getArgumentTransforms(t4).isEmpty, isTrue);
  });

  test('isNormalizedMapping', () {
    final t1 = parseTransform('a + a => b + c');
    final m = findSmallestTransforms(t1);
    expect(isNormalizedMapping(m), isFalse);
  });

  test('substituteTransform', () {
    final input = parse('a + r^2*r^5');
    final rule = parseTransform('#x^#a*#x^#b => #x^(#a+#b)');
    final result = substituteTransform(input, 2, rule);
    final evaluated = tryEvaluateExpr(result, canEval, eval);
    expect(evaluated, equals(parse('a + r^7')));
  });
}
