// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

library expr.test.lemma_test;

import 'dart:io';
import 'package:test/test.dart';
import 'package:expr/expr.dart';
import 'package:expr/testenv.dart';

import 'misc/helpers.dart';

// The test will construct a number of proofs by reading the formulas
// provided by `database/formulas.txt` and the proofs in `database/proofs`.
const formulasFile = 'test/data/formulas.txt';
const proofsDir = 'test/data/proofs';

/// Line in a proof document
class ProofRow {
  final Expr expr;
  final String flags;
  ProofRow(this.expr, this.flags);
}

void main() {
  // All formulas
  final formulaModes = new Map<Transform, CollapseMode>();
  final formulas = new File(formulasFile)
      .readAsLinesSync()
      .where((l) => l.isNotEmpty)
      .map((line) {
    final retain = line.startsWith('!');
    final t = parseTransform(line.substring(retain ? 1 : 0));
    formulaModes[t] = retain ? CollapseMode.adjacent : null;
    return new Formula([], t);
  }).toList();

  // Read test proof input
  //
  // Note that negated integers are not initially evaluated. This has
  // consequences for which formulas can be applied to negative integers.
  // Negative integers can of course be created by evaluation. As a solution we
  // could consider wrapping negative integers in a negate function after
  // evaluation.
  final proofs = new Directory(proofsDir).listSync().map((fse) {
    return new File(fse.path)
        .readAsLinesSync()
        .where((line) => !line.startsWith('#'))
        .map((line) {
      return new ProofRow(parse(line.substring(4)), line.substring(0, 3));
    }).toList();
  });

  test('proveLemma', () {
    // Load for pretty-printing errors.
    loadTemplates(defaultTemplates);

    // Loop through test proofs.
    for (final steps in proofs) {
      // Reconstruct proofs.
      Expr previousExpr;
      final rewrites = new List<Rewrite>();
      for (final step in steps.where((s) => s.flags[0] == '>')) {
        if (previousExpr != null) {
          final t = new Transform(previousExpr, step.expr);
          final l = proveLemma(t, canEval, eval, canRearrange, canRedistribute,
              formulas, getExpansion, getAssumption);

          // Throw a readable exception when no proof can be found.
          if (l == null) {
            final str1 = easyPrintExpr(t.source);
            final str2 = easyPrintExpr(t.target);
            throw new Exception('unable to prove transform: $str1 => $str2');
          } else {
            rewrites.addAll(l.rewrites);
          }
        }
        previousExpr = step.expr;
      }

      // Check expected number of rewrites.
      expect(rewrites.length, equals(steps.length - 1));

      // Check each step.
      for (var i = 0; i < rewrites.length; i++) {
        final input = steps[i].expr;
        final output = applyRewrite(input, rewrites[i], eval);
        expect(output, equals(steps[i + 1].expr));
      }

      // Split into parts using different collapse modes for substitute.
      var flagIndex = 0;
      final lemma = new Lemma.from(steps.first.expr, rewrites, eval);
      for (final subsMode in [CollapseMode.append, CollapseMode.prepend]) {
        // Compute splits.
        final parts = splitLemma(lemma, eval, canRedistribute.keys.toSet(),
            substitute: (s) => formulaModes[s.formula.proposition] ?? subsMode);
        final statements = parts.map((p) => p.statement).toList();

        // Get expected splits.
        flagIndex++;
        final splits = steps
            .where((s) => s.flags[flagIndex] == '-')
            .map((s) => s.expr)
            .toList();
        final expected = new List<Transform>.generate(
            splits.length - 1, (i) => new Transform(splits[i], splits[i + 1]));

        // Compare pretty-printed strings for easy debugging.
        final pr = easyPrintExpr;
        String pt(Transform t) => '${pr(t.source)} => ${pr(t.target)}';
        expect(statements.map(pt).toList(), equals(expected.map(pt).toList()));
        expect(statements, equals(expected));
      }
    }

    clearTemplates();
  });

  test('proveLemma with backward premises', () {
    final moreFormulas = formulas.toList();
    final formula = new Formula(
        [parseTransform('#a => True'), parseTransform('#b => True')],
        parseTransform('and(#a, #b) => True'));
    final proof1 = new Formula([], parseTransform('x => True'));
    final proof2 = new Formula([], parseTransform('y => True'));
    moreFormulas.addAll([formula, proof1, proof2]);

    final t = parseTransform('and(x, y) => True');
    final lemma = proveLemma(t, canEval, eval, canRearrange, canRedistribute,
        moreFormulas, getExpansion, getAssumption);

    final st = lemma.rewrites.single.statement;
    final pr = lemma.rewrites.single.statement.premiseProofs;
    expect(st.formula, equals(formula));
    expect(pr[0].rewrites.single.statement.formula, equals(proof1));
    expect(pr[1].rewrites.single.statement.formula, equals(proof2));
  });

  test('proveLemma with assumption', () {
    TrueStatement<String> _getAssumption(t) => new TrueStatement(
        new Formula<String>([], t, metadata: 'assumption'), []);

    final t = parseTransform('a + b => a + c');
    final l1 = proveLemma(t, canEval, eval, canRearrange, canRedistribute, [],
        getExpansion, getAssumption);
    final l2 = proveLemma(t, canEval, eval, canRearrange, canRedistribute, [],
        getExpansion, _getAssumption);

    expect(l1, isNull);
    expect(l2.rewrites.single.statement.formula.metadata, equals('assumption'));
  });

  test('applyRewrite with forward premises', () {
    final formula = new Formula(
        [parseTransform('#a => #b'), parseTransform('#b => #c')],
        parseTransform('#a => #c'));

    final t1 = parseTransform('a1 => b1');
    final t2 = parseTransform('b1 => c1');
    final p1 = new Lemma.from(parse('a1'), [rewriteSubstitute(0, t1)], eval);
    final p2 = new Lemma.from(parse('b1'), [rewriteSubstitute(0, t2)], eval);
    final st = new TrueStatement(formula, [p1, p2]);

    final result = applyRewrite(parse('a1'), rewriteSubstitute(0, st), eval);
    expect(result, equals(parse('c1')));
  });
}
