// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

library expr.test.print_test;

import 'package:test/test.dart';
import 'package:expr/expr.dart';
import 'package:expr/testenv.dart';

final latexFunctionTemplates = {
  // Operators
  '+': r'${.0}+${1(+).}',
  '-': r'${.0}-${1(+).}',
  '*': r'${.0(+):}${:1(*).}',
  '/': r'\frac{${0}}{${1}}',
  '^': r'{${.0(*)}}^{${1}}',
  '~': r'$!-${0(^).}',
  '!': r'${.0(~)}!',
  '_': r'{${.0(!)}}_{${1}}',

  // Other functions
  'diff': r'\frac{\partial}{\partial${:0(+)}}${:1(+).}',
  'int': r'\int_{${0}}^{${1}}${2}~\text{d}${:3(+).}',
  'sin': r'\sin${:0(+).}',
  'delta': r'\Delta${:0(+).}',
  'celsius': r'${.0(^)}^\circ',
  'atleast': r'${0(+)}^+$!'
};

final latexPrintTests = {
  '1 * 2': r'1\left(2\right)',
  '2 * 10^a': r'2\left({10}^{a}\right)',
  '(a - -5) ^ (b * c)': r'{\left(a--5\right)}^{b c}',
  '(a / -5) ^ (b * c)': r'{\left(\frac{a}{-5}\right)}^{b c}',
  'a*(1/a)': r'a\frac{1}{a}',
  '-(1 ^ 2)': r'-\left({1}^{2}\right)',
  '(-1) ^ 2': r'{-1}^{2}',
  '-1 ^ 2': r'{-1}^{2}',
  '---a': r'---a',
  'a + b + -c': r'a+b+-c',
  'a + b * -c': r'a+b\left(-c\right)',
  'celsius(10)': r'10^\circ',
  'celsius(a+10)': r'\left(a+10\right)^\circ',
  'celsius(10)*2': r'10^\circ\left(2\right)',
  'a*celsius(10)': r'a\left(10^\circ\right)',
  '2*celsius(10)': r'2\left(10^\circ\right)',
  'a*celsius(-10)': r'a\left(-10^\circ\right)',
  '2*celsius(-10)': r'2\left(-10^\circ\right)',
  'delta(2x)+sin(2a)^2': r'\Delta2x+{\left(\sin2a\right)}^{2}',
  'delta(1+x)+sin(a)^2': r'\Delta\left(1+x\right)+{\left(\sin a\right)}^{2}',
  '2*sin(a*b)': r'2\sin a b',
  'sin(a*b)*2': r'\left(\sin a b\right)2',
  'sin(a)^2': r'{\left(\sin a\right)}^{2}',
  'sin(a^2)': r'\sin{a}^{2}',
  'atleast(10)': r'10^+',
  'atleast(10)+1': r'10^++1',
  'atleast(10)*1': r'10^+\left(1\right)',

  // Integrals
  '2*int(0,1,x^2+2x+1,1/x)': r'2\int_{0}^{1}{x}^{2}+2x+1~\text{d}\frac{1}{x}',
  'int(0,1,x^2+2x+1,1/x)*2':
      r'\left(\int_{0}^{1}{x}^{2}+2x+1~\text{d}\frac{1}{x}\right)2',
  'int(0,1,x^2+2x+1,1/x+1)*2': r'\left(\int_{0}^{1}{x}^{2}+2x+1~\text{d}'
      r'\left(\frac{1}{x}+1\right)\right)2',

  // Derivatives
  'diff(2x, x+2)': r'\frac{\partial}{\partial2x}\left(x+2\right)',
  'diff(x, -a)': r'\frac{\partial}{\partial x}\left(-a\right)',
  'diff(x, -1)': r'\frac{\partial}{\partial x}\left(-1\right)',
  'diff(x, -1*a)': r'\frac{\partial}{\partial x}\left(-1a\right)',
  'diff(-1, -1)*2': r'\left(\frac{\partial}{\partial\left(-1\right)}'
      r'\left(-1\right)\right)2',
  'diff(-1, -1)^2': r'{\left(\frac{\partial}{\partial\left(-1\right)}'
      r'\left(-1\right)\right)}^{2}',
  'diff(x+1, x+2)*2': r'\left(\frac{\partial}{\partial\left(x+1\right)}'
      r'\left(x+2\right)\right)2'
};

/// LaTeX specific variant of [getTemplate].
PrintTemplate getLaTeXTemplate(ExprObject obj) {
  return getTemplate(obj, r'\left(', r'\right)', r'\,,', (str) => '\\{$str}');
}

void main() {
  test('printExpr', () {
    loadTemplates(latexFunctionTemplates);

    // Create operator map.
    final operatorId2Pre = new Map<int, int>.fromIterable(operators,
        key: (op) => op.functionId, value: (op) => op.precedenceLevel);

    // Run each specified case.
    latexPrintTests.forEach((input, expectedOutput) {
      // Get print output and compare with expected value.
      final actualOutput = printExpr(parse(input), operatorId2Pre, r'\left(',
          r'\right)', getLaTeXTemplate, insertLaTeXSeparator);
      expect(actualOutput, equals(expectedOutput));
    });

    clearTemplates();
  });

  test('checkPrintTemplateArguments', () {
    PrintTemplate parse(String str) => parsePrintTemplate(str, operators);
    final valid = [r'\alpha', r'\frac{${0}}{${1}}', r'\frac{${0}}{${0}}'];
    final invalid = [r'\frac{${1}}{${2}}', r'\frac{${0}}{${2}}'];

    for (final str in valid) {
      expect(checkPrintTemplateArguments(parse(str)), isTrue);
    }
    for (final str in invalid) {
      expect(checkPrintTemplateArguments(parse(str)), isFalse);
    }
  });
}
