// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

library expr.test.throw_test;

import 'package:test/test.dart';
import 'package:expr/expr.dart';
import 'package:expr/testenv.dart';

import 'misc/helpers.dart';
import 'misc/matchers.dart';

void main() {
  group('core', testCoreExceptions);
  group('parse', testParseExceptions);
  group('print', testPrintExceptions);
  group('lemma', testLemmaExceptions);
}

/// All tests associated to core functions
void testCoreExceptions() {
  test('compareExprObjects', () {
    final t = parseTransform('f(1, 2, 3) => f(1, 2)');
    expect(() => compareExprObjects(t.source.first, t.target.first),
        thrws('incompatible argument count between a and b'));
  });

  test('evaluateExpr', () {
    expect(() => evaluateExpr(parse('1 + #a'), eval),
        thrws('input cannot contain patterns'));

    final incomplete = new Expr(parse('1 + 1').sublist(0, 1));
    expect(() => evaluateExpr(incomplete, eval),
        thrws('incomplete input expression'));
  });

  test('rearrangeExpr', () {
    final r = new Rearrangement(0, [-1, 0, 1]);
    expect(() => rearrangeExpr(parse('1'), r),
        thrws('rearrangement target is not a function'));
  });

  test('findRedistributions', () {
    final expr = parse('f(g(a, b, c))');
    final key = new DistFnKey(fnId('f'), fnId('g'));
    expect(() => unwindExprDist(expr, key, [0]),
        thrws('illegal g function argument count'));
  });

  test('redistributeExpr', () {
    final key = new DistFnKey(fnId('*'), fnId('+'));
    final node1a = new RedistrNode(0, null, null, null, null);
    final node1b = new RedistrNode(1, null, null, null, null);
    final node1 = new RedistrNode(-1, null, node1a, node1b, [1, 0]);
    final node2 = new RedistrNode(-1, null, node1a, node1b, [1]);

    final r1 = new Redistribution(key, [0, 1], node1);
    final r2 = new Redistribution(key, [0, 1], node2);

    expect(() => redistributeExpr(parse('a*b(c) + c*d'), r1),
        thrws('invalid position'));
    expect(() => redistributeExpr(parse('a*b(c) + c*d'), r2),
        thrws('asymmetric merge'));
  });

  test('mapExpr', () {
    // This is quite a peculiar case.
    final input = parse('#f(#x + 1, #y + 2)');
    final mapping = new Mapping();
    mapping[parse('#x')] = [parse('z')].toSet();
    mapping[parse('#y')] = [parse('z')].toSet();
    mapping[parse('#f(#x, #y)')] = [parse('a(z)b(z)')].toSet();

    expect(() => mapExpr(input, mapping),
        thrws('conflict in pattern to pattern mapping'));
  });

  test('substituteTransform', () {
    final input1 = parse('r^3');
    final rule1 = parseTransform('#x^2 => #x#x');
    expect(() => substituteTransform(input1, 0, rule1),
        thrws('non-trivial transforms in mapping'));

    final input2 = parse('r^3+r^1');
    final rule2 = parseTransform('#x^2+#x^2 => #x#x+#x#x');
    expect(() => substituteTransform(input2, 0, rule2),
        thrws('conflicts in mapping normalization'));
  });
}

/// All tests associated with parse functions
void testParseExceptions() {
  test('parseExpr', () {
    expect(() => parseExpr('', [], null, ' '), thrws('illegal pattern symbol'));
    expect(() => parseExpr('', [], null, ''), thrws('illegal pattern symbol'));
    expect(() => parseExpr('', [], null), thrws('input cannot be empty'));

    expect(() => parseExpr(')', operators, getId),
        thrws('mismatched parentheses'));
    expect(() => parseExpr('a + b)', operators, getId),
        thrws('mismatched parentheses'));

    expect(() => parseExpr(',', operators, getId),
        thrws('argument separator outside function'));
    expect(() => parseExpr('a + b,', operators, getId),
        thrws('argument separator outside function'));
    expect(() => parseExpr('(a + b,', operators, getId),
        thrws('argument separators outside function'));

    // This one is more tricky to reproduce. We start with a negation operator
    // and end the string before adding any arguments.
    expect(() => parseExpr('~', operators, getId),
        thrws('stack function arguments are not in the output queue'));

    // A normal run to test if [parseExpr] can find the multiplyId by itself.
    expect(parseExpr('a * b', operators, getId), equals(parse('a * b')));
  });
}

/// All tests associated with print functions
void testPrintExceptions() {
  test('printTemplate', () {
    final template = parsePrintTemplate(r'${1}', []);
    final parent = new ExprObject(ExprType.function, 0, 1);
    expect(
        () => printTemplate(template, parent, [], [], {}, '', '', (a, b) => ''),
        thrws('bad template; invalid argument index'));
  });
}

/// All tests associated with lemma functions
void testLemmaExceptions() {
  test('applyRewrite', () {
    expect(() => applyRewrite(parse('a'), new Rewrite(0, null), eval),
        thrws('unable to process rewrite instruction'));

    // Try statement with unnormalized proof.
    final formula = new Formula(
        [parseTransform('#a => #b'), parseTransform('#b => #c')],
        parseTransform('#a => #c'));
    final lemma1 = new Lemma.from(
        parse('a1'), [rewriteSubstitute(0, parseTransform('a1 => b1'))], eval);
    final lemma2 = new Lemma.from(
        parse('b1'), [rewriteSubstitute(0, parseTransform('b1 => c1'))], eval);
    final lemma3 = new Lemma.from(
        parse('b2'), [rewriteSubstitute(0, parseTransform('b2 => c2'))], eval);

    final statement1 = new TrueStatement(formula, [lemma1, lemma2]);
    final statement2 = new TrueStatement(formula, [lemma1, lemma3]);
    final rewrite1 = rewriteSubstitute(0, statement1);
    final rewrite2 = rewriteSubstitute(0, statement2);

    expect(applyRewrite(parse('a1'), rewrite1, eval), equals(parse('c1')));
    expect(() => applyRewrite(parse('a1'), rewrite2, eval),
        thrws('statement has unnormalized proof'));
  });
}
